"""Provides the adRegion display tools."""

from __future__ import annotations

import time
from typing import TYPE_CHECKING, cast

import numpy as np
from adafruit_rgb_display.st7789 import ST7789
from fake import Fake

from ubo_app.logger import logger
from ubo_app.store.services.display import (
    DisplayCompressedRenderEvent,
    DisplayRenderEvent,
)
from ubo_app.utils import IS_RPI, IS_TEST_ENV
from ubo_app.utils.eeprom import get_eeprom_data

if TYPE_CHECKING:
    from collections.abc import Callable

    from headless_kivy.config import Region


from ubo_app.constants import DISPLAY_BAUDRATE, HEIGHT, WIDTH


class Display:
    """Display class."""

    def __init__(self: Display) -> None:
        """Initialize the display."""
        self.cs_pin = None
        self.dc_pin = None
        self.reset_pin = None
        self.backlight_pin = None
        self.spi = None
        self.display = None
        if IS_RPI:
            eeprom_data = get_eeprom_data()

            if (
                (lcd := eeprom_data.get('lcd')) is not None
                and lcd.get('model') == 'st7789'
            ):
                logger.debug('LCD display found.')
                import board
                import digitalio

                self.cs_pin = digitalio.DigitalInOut(board.CE0)
                self.dc_pin = digitalio.DigitalInOut(board.D25)
                self.spi = board.SPI()
                self.display = ST7789(
                    self.spi,
                    height=HEIGHT,
                    width=WIDTH,
                    y_offset=80,
                    x_offset=0,
                    cs=self.cs_pin,
                    dc=self.dc_pin,
                    baudrate=DISPLAY_BAUDRATE,
                )

                self.backlight_pin = digitalio.DigitalInOut(board.D26)
                self.backlight_pin.switch_to_output()
                self.backlight_pin.value = True
        else:
            logger.debug('No physical display found.')
            self.display = cast('ST7789', Fake())

    def turn_off(self: Display) -> None:
        """Turn off the display and free resources."""
        if self.display:
            render = self.display._block  # noqa: SLF001
            self.display = None
            self.render_blank(render)
            del render

            if IS_RPI:
                from adafruit_blinka.agnostic import detector

                if detector.board.any_raspberry_pi_5_board:
                    import lgpio  # pyright: ignore [reportMissingModuleSource]
                    from adafruit_blinka.microcontroller.generic_linux import lgpio_pin

                    lgpio.gpiochip_close(lgpio_pin.CHIP)
                else:
                    import board
                    from RPi import GPIO  # pyright: ignore [reportMissingModuleSource]

                    if board.CE0.id:
                        GPIO.cleanup(board.CE0.id)
                    if board.D25.id:
                        GPIO.cleanup(board.D25.id)
                    if board.D26.id:
                        GPIO.cleanup(board.D26.id)

    def set_backlight(self: Display, enabled: bool) -> None:  # noqa: FBT001
        """Control backlight state."""
        if IS_RPI and self.backlight_pin is not None:
            self.backlight_pin.value = enabled
            logger.debug(
                'Backlight state changed',
                extra={'enabled': enabled},
            )

    def render_blank(self: Display, render_function: Callable | None = None) -> None:
        """Render a blank screen."""
        if IS_RPI:
            if not render_function and self.display is not None:
                render_function = self.display._block  # noqa: SLF001
            if render_function:
                render_function(
                    0,
                    0,
                    WIDTH - 1,
                    HEIGHT - 1,
                    b'\x00\x00' * WIDTH * HEIGHT,
                )
                time.sleep(0.2)
                render_function(
                    0,
                    0,
                    WIDTH - 1,
                    HEIGHT - 1,
                    b'\x00\x00' * WIDTH * HEIGHT,
                )

    def render_block(
        self: Display,
        *,
        rectangle: tuple[int, int, int, int],
        data_bytes: bytes,
        bypass_pause: bool = False,
    ) -> None:
        """Block the display."""
        from ubo_app.store.main import store

        @store.with_state(
            lambda state: state.display.is_paused
            if hasattr(state, 'display')
            else False,
        )
        def render(is_paused: bool) -> None:  # noqa: FBT001
            if self.display is not None and (not is_paused or bypass_pause):
                self.display._block(*rectangle, data_bytes)  # noqa: SLF001

        render()


display = Display()


def render_on_display(*, regions: list[Region]) -> None:
    """Transfer data to the display via SPI controller."""
    from ubo_app.store.main import store


    for region in regions:
        rectangle = region['rectangle']
        data = region['data'].astype(np.uint16)
        color = (
            ((data[:, :, 0] & 0xF8) << 8)
            | ((data[:, :, 1] & 0xFC) << 3)
            | (data[:, :, 2] >> 3)
        ).copy()
        data_bytes = (
            color.astype(np.uint16).view(np.uint8).reshape(-1, 2)[:, ::-1].tobytes()
        )
        display.render_block(
            rectangle=(
                rectangle[1],
                rectangle[0],
                rectangle[3] - 1,
                rectangle[2] - 1,
            ),
            data_bytes=data_bytes,
        )

    from kivy.metrics import dp

    density = dp(1)

    def generate_render_actions(
        region: Region,
    ) -> tuple[DisplayRenderEvent, DisplayCompressedRenderEvent]:
        import zlib

        data = region['data'].tobytes()
        compressor = zlib.compressobj(wbits=-zlib.MAX_WBITS)
        compressed_data = compressor.compress(data) + compressor.flush()
        timestamp = time.time()
        return (
            DisplayRenderEvent(
                timestamp=timestamp,
                data=data,
                rectangle=region['rectangle'],
                density=density,
            ),
            DisplayCompressedRenderEvent(
                timestamp=timestamp,
                compressed_data=compressed_data,
                rectangle=region['rectangle'],
                density=density,
            ),
        )

    if not IS_TEST_ENV:
        from ubo_app.store.main import store

        store._dispatch(  # noqa: SLF001
            [event for region in regions for event in generate_render_actions(region)],
        )


splash_screen = None
if splash_screen:
    display.render_block(
        rectangle=(0, 0, WIDTH - 1, HEIGHT - 1),
        data_bytes=splash_screen,
    )
else:
    display.render_blank()



# ruff: noqa: D100, D103
from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, ParamSpec

from immutable import Immutable
from typing_extensions import TypeVar

from ubo_app.utils.service import get_coroutine_runner

if TYPE_CHECKING:
    from asyncio import Handle
    from collections.abc import Callable, Coroutine

    from redux.basic_types import TaskCreatorCallback

    from ubo_app.utils.types import CoroutineRunner


tasks: list[Handle] = []


def create_task(
    task: Coroutine,
    callback: TaskCreatorCallback | None = None,
    coroutine_runner: CoroutineRunner | None = None,
    name: str | None = None,
) -> Handle:
    def callback_(task: asyncio.Task) -> None:
        if callback:
            callback(task)

    handle: Handle | None = None
    signal = asyncio.Event()

    async def wrapper() -> None:
        try:
            await task
        finally:
            await signal.wait()
            if handle in tasks:
                tasks.remove(handle)

    result = wrapper()
    result.__name__ = f'task_wrapper_coroutine:{task.__name__}'
    result.__qualname__ = f'task_wrapper_coroutine:{task.__qualname__}'

    if coroutine_runner is None:
        coroutine_runner = get_coroutine_runner()

    handle = (
        coroutine_runner(result, callback_, name=name)
        if callback
        else coroutine_runner(result, name=name)
    )

    tasks.append(handle)
    signal.set()
    return handle


T = TypeVar('T', infer_variance=True)
T_params = ParamSpec('T_params')


class ToThreadOptions(Immutable):
    """Options for the to_thread function."""

    coroutine_runner: CoroutineRunner | None = None
    callback: TaskCreatorCallback | None = None
    name: str | None = None


def to_thread(
    task: Callable[T_params, T],
    options: ToThreadOptions | None = None,
    *args: T_params.args,
    **kwargs: T_params.kwargs,
) -> Handle:
    coroutine_runner = get_coroutine_runner()
    if options is None:
        options = ToThreadOptions()

    if options.coroutine_runner is not None:
        coroutine_runner = options.coroutine_runner

    return coroutine_runner(
        asyncio.to_thread(task, *args, **kwargs),
        callback=options.callback,
        name=options.name,
    )



"""Provides `IS_RPI` and `IS_TEST_ENV` constants."""

import contextlib
import os
import sys
from pathlib import Path

IS_RPI = Path('/etc/rpi-issue').exists()
IS_RPI_4 = False
with contextlib.suppress(Exception):
    IS_RPI_4 = IS_RPI and Path('/proc/device-tree/model').read_text().startswith(
        'Raspberry Pi 4',
    )
# Check for pytest in argv (local tests) or UBO_TEST_ENV env var (device tests)
IS_TEST_ENV = any('pytest' in arg.lower() for arg in sys.argv) or bool(
    os.environ.get('UBO_TEST_ENV'),
)

from ubo_app.utils.eeprom import (  # noqa: E402
    UNAVAILABLE_SERIAL_NUMBER,
    read_serial_number,
)

IS_UBO_POD = read_serial_number() is not UNAVAILABLE_SERIAL_NUMBER



# ruff: noqa: D100, D101
from __future__ import annotations

from dataclasses import field
from enum import StrEnum, auto

from immutable import Immutable
from redux import BaseAction, BaseEvent

UPDATE_MANAGER_NOTIFICATION_ID = 'ubo:update_manager'
UPDATE_MANAGER_SECOND_PHASE_NOTIFICATION_ID = 'ubo:update_manager:phase-2'


class UpdateManagerAction(BaseAction): ...


class UpdateManagerSetVersionsAction(UpdateManagerAction):
    flash_notification: bool
    current_version: str
    base_image_variant: str
    latest_version: str
    recent_versions: list[str] | None = None


class UpdateManagerRequestCheckAction(UpdateManagerAction): ...


class UpdateManagerReportFailedCheckAction(UpdateManagerAction): ...


class UpdateManagerRequestUpdateAction(UpdateManagerAction):
    version: str


class UpdateManagerEvent(BaseEvent): ...


class UpdateManagerCheckEvent(UpdateManagerEvent): ...


class UpdateManagerUpdateEvent(UpdateManagerEvent):
    version: str | None


class UpdateStatus(StrEnum):
    """Update status enum."""

    CHECKING = auto()
    FAILED_TO_CHECK = auto()
    UP_TO_DATE = auto()
    OUTDATED = auto()
    UPDATING = auto()


class UpdateManagerState(Immutable):
    """Version store."""

    current_version: str | None = None
    base_image_variant: str | None = None
    latest_version: str | None = None
    update_status: UpdateStatus = UpdateStatus.CHECKING
    recent_versions: list[str] = field(default_factory=list)



"""gRPC service that implements the Store service."""

from __future__ import annotations

from ubo_app.logger import logger
from ubo_app.utils import secrets

from ubo_bindings.secrets.v1 import (
    QuerySecretRequest,
    QuerySecretResponse,
    SecretsServiceBase,
)


class SecretsService(SecretsServiceBase):
    """gRPC service class that implements the Store service."""

    async def query_secret(
        self,
        query_secret_request: QuerySecretRequest,
    ) -> QuerySecretResponse:
        """Query a secret from the store."""
        logger.debug(
            'Querying secret',
            extra={'key': query_secret_request.key},
        )

        secret = secrets.read_secret(query_secret_request.key)
        if secret is None:
            return QuerySecretResponse(error='Secret not found')

        return QuerySecretResponse(value=secret)



"""Engines registry."""

from ubo_app.engines.abstraction.ai_provider_mixin import AIProviderMixin
from ubo_app.engines.assemblyai import AssemblyAIEngine
from ubo_app.engines.cerebras import CerebrasEngine
from ubo_app.engines.deepgram import DeepgramEngine
from ubo_app.engines.elevenlabs import ElevenLabsEngine
from ubo_app.engines.google import GoogleEngine
from ubo_app.engines.google_cloud import GoogleCloudEngine
from ubo_app.engines.grok import GrokEngine
from ubo_app.engines.ollama import OllamaEngine
from ubo_app.engines.ollama_onprem import OllamaOnPremEngine
from ubo_app.engines.openai import OpenAIEngine
from ubo_app.engines.piper import PiperEngine
from ubo_app.engines.rime import RimeEngine
from ubo_app.engines.vosk import VoskEngine
from ubo_app.store.services.assistant import (
    AssistantImageGeneratorName,
    AssistantLLMName,
    AssistantSTTName,
    AssistantTTSName,
)

STT_ENGINES: dict[AssistantSTTName, AIProviderMixin] = {
    AssistantSTTName.VOSK: VoskEngine(),
    AssistantSTTName.GOOGLE: GoogleCloudEngine(label='Google (continuous)'),
    AssistantSTTName.GOOGLE_SEGMENTED: GoogleCloudEngine(label='Google (segmented)'),
    AssistantSTTName.OPENAI: OpenAIEngine(),
    AssistantSTTName.DEEPGRAM: DeepgramEngine(),
    AssistantSTTName.ASSEMBLYAI: AssemblyAIEngine(),
}

LLM_ENGINES: dict[AssistantLLMName, AIProviderMixin] = {
    AssistantLLMName.OLLAMA: OllamaEngine(),
    AssistantLLMName.OLLAMA_ONPREM: OllamaOnPremEngine(),
    AssistantLLMName.GOOGLE: GoogleCloudEngine(label='Google Vertex'),
    AssistantLLMName.GROK: GrokEngine(),
    AssistantLLMName.CEREBRAS: CerebrasEngine(),
    AssistantLLMName.OPENAI: OpenAIEngine(),
}

TTS_ENGINES: dict[AssistantTTSName, AIProviderMixin] = {
    AssistantTTSName.PIPER: PiperEngine(),
    AssistantTTSName.GOOGLE: GoogleCloudEngine(label='Google'),
    AssistantTTSName.OPENAI: OpenAIEngine(),
    AssistantTTSName.ELEVENLABS: ElevenLabsEngine(),
    AssistantTTSName.RIME: RimeEngine(),
}

IMAGE_GENERATOR_ENGINES: dict[
    AssistantImageGeneratorName,
    AIProviderMixin,
] = {
    AssistantImageGeneratorName.GOOGLE: GoogleEngine(),
    AssistantImageGeneratorName.OPENAI: OpenAIEngine(),
}



# ruff: noqa: D100, D103
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ubo_handle import register


def setup() -> None:
    from setup import init_service

    init_service()


register(
    service_id='ethernet',
    label='Ethernet',
    setup=setup,
)



"""Defines the color constants used in the application."""

from ubo_gui.constants import (
    DANGER_COLOR,
    INFO_COLOR,
    PRIMARY_COLOR,
    SECONDARY_COLOR,
    SECONDARY_COLOR_LIGHT,
    SUCCESS_COLOR,
    WARNING_COLOR,
)

RUNNING_COLOR = '#008000'
STOPPED_COLOR = '#ffff00'

__all__ = [
    'DANGER_COLOR',
    'INFO_COLOR',
    'PRIMARY_COLOR',
    'RUNNING_COLOR',
    'SECONDARY_COLOR',
    'SECONDARY_COLOR_LIGHT',
    'STOPPED_COLOR',
    'SUCCESS_COLOR',
    'WARNING_COLOR',
]



"""Definitions for speech recognition service actions, events and state."""

from __future__ import annotations

from dataclasses import field
from enum import StrEnum
from typing import TYPE_CHECKING

from immutable import Immutable
from redux import BaseAction, BaseEvent

if TYPE_CHECKING:
    from collections.abc import Sequence

    from ubo_app.store.main import UboAction

from ubo_app.utils.persistent_store import read_from_persistent_store


class SpeechRecognitionAction(BaseAction):
    """Base class for speech recognition actions."""


class SpeechRecognitionSetSelectedEngineAction(SpeechRecognitionAction):
    """Action to set the selected speech recognition engine."""

    engine_name: SpeechRecognitionEngineName | None


class SpeechRecognitionSetIsIntentsActiveAction(SpeechRecognitionAction):
    """Action to set the active state of the voice intents listener."""

    is_active: bool


class SpeechRecognitionSetIsAssistantActiveAction(SpeechRecognitionAction):
    """Action to set the active state of the voice assistant listener."""

    is_active: bool


class SpeechRecognitionIntent(Immutable):
    """Intent for speech recognition service."""

    phrase: str | Sequence[str]
    action: UboAction | Sequence[UboAction]


class SpeechRecognitionReportWakeWordDetectionAction(SpeechRecognitionAction):
    """Action to report wake word detection."""

    wake_word: str


class SpeechRecognitionReportIntentDetectionAction(SpeechRecognitionAction):
    """Action to report intent detection."""

    intent: SpeechRecognitionIntent


class SpeechRecognitionReportSpeechAction(SpeechRecognitionAction):
    """Action to report speech raw audio and recognized text."""

    engine_name: SpeechRecognitionEngineName
    text: str
    audio: bytes


class SpeechRecognitionEvent(BaseEvent):
    """Base class for speech recognition events."""


class SpeechRecognitionReportTextEvent(SpeechRecognitionEvent):
    """Event to report stream of recognized text."""

    timestamp: float
    text: str


class SpeechRecognitionStatus(StrEnum):
    """State for speech recognition service."""

    IDLE = 'idle'
    INTENTS_WAITING = 'intents_waiting'
    ASSISTANT_WAITING = 'assistant_waiting'


class SpeechRecognitionEngineName(StrEnum):
    """Available speech recognition engines."""

    VOSK = 'vosk'
    GOOGLE = 'google_cloud'


class SpeechRecognitionState(Immutable):
    """State for speech recognition service."""

    selected_engine: SpeechRecognitionEngineName | None = field(
        default=read_from_persistent_store(
            'speech_recognition:selected_engine',
            mapper=lambda value: SpeechRecognitionEngineName(value)
            if value in SpeechRecognitionEngineName.__members__.values()
            else SpeechRecognitionEngineName.VOSK,
            default=SpeechRecognitionEngineName.VOSK,
        ),
    )
    intents: list[SpeechRecognitionIntent]
    is_intents_active: bool = field(
        default=read_from_persistent_store(
            'speech_recognition:is_intents_active',
            default=True,
        ),
    )
    is_assistant_active: bool = field(
        default=read_from_persistent_store(
            'speech_recognition:is_assistant_active',
            default=False,
        ),
    )
    status: SpeechRecognitionStatus = SpeechRecognitionStatus.IDLE



# ruff: noqa: D100, D103
from __future__ import annotations

import pathlib
from typing import TYPE_CHECKING

from commands import (
    check_is_active,
    check_status,
    install_rpi_connect,
    sign_out,
    start_service,
    stop_service,
    uninstall_rpi_connect,
)
from kivy.lang.builder import Builder
from sign_in_page import SignInPage
from ubo_gui.menu.types import ActionItem, ApplicationItem, HeadedMenu

from ubo_app.store.core.types import RegisterSettingAppAction, SettingsCategory
from ubo_app.store.main import store
from ubo_app.store.ubo_actions import UboApplicationItem, register_application
from ubo_app.utils.async_ import create_task
from ubo_app.utils.gui import UboPageWidget

if TYPE_CHECKING:
    from ubo_app.store.services.rpi_connect import (
        RPiConnectState,
        RPiConnectStatus,
    )


class _RPiConnectQRCodePage(UboPageWidget): ...


register_application(
    application=_RPiConnectQRCodePage,
    application_id='rpi-connect:qrcode-page',
)
register_application(
    application=SignInPage,
    application_id='rpi-connect:signin-page',
)


def status_based_actions(
    status: RPiConnectStatus,
) -> list[ActionItem | ApplicationItem]:
    actions = []

    if (
        status.screen_sharing_sessions is not None
        or status.remote_shell_sessions is not None
    ):
        actions.append(
            UboApplicationItem(
                label='Show URL',
                icon='󰐲',
                application_id='rpi-connect:qrcode-page',
            ),
        )
    return actions


def login_actions(*, is_signed_in: bool | None) -> list[ActionItem | ApplicationItem]:
    actions = []
    if is_signed_in:
        actions.append(
            ActionItem(label='Sign out', icon='󰍃', action=sign_out),
        )
    elif is_signed_in is False:
        actions.append(
            UboApplicationItem(
                label='Sign in',
                icon='󰍂',
                application_id='rpi-connect:signin-page',
            ),
        )
    return actions


@store.autorun(lambda state: state.rpi_connect)
def actions(state: RPiConnectState) -> list[ActionItem | ApplicationItem]:
    actions = []
    if not state.is_downloading:
        if state.is_installed:
            if state.status:
                actions.extend(status_based_actions(state.status))
            actions.extend(login_actions(is_signed_in=state.is_signed_in))
            actions.append(
                ActionItem(
                    label='Stop' if state.is_active else 'Start',
                    icon='󰓛' if state.is_active else '󰐊',
                    action=stop_service if state.is_active else start_service,
                ),
            )

        if state.is_installed is not None:
            actions.append(
                ActionItem(
                    label='Uninstall RPi-Connect'
                    if state.is_installed
                    else 'Install RPi-Connect',
                    icon='󰇚',
                    action=uninstall_rpi_connect
                    if state.is_installed
                    else install_rpi_connect,
                ),
            )
    return actions


@store.autorun(lambda state: state.rpi_connect)
def status(state: RPiConnectState) -> str:
    if state.status:
        status = 'Screen sharing: '
        if state.status.screen_sharing_sessions is None:
            status += 'unavailable'
        else:
            status += f'{state.status.screen_sharing_sessions} sessions'
        status += '\nRemote shell: '
        if state.status.remote_shell_sessions is None:
            status += 'unavailable'
        else:
            status += f'{state.status.remote_shell_sessions} sessions'
    elif state.is_downloading:
        status = 'Downloading...'
    elif state.is_installed is None:
        status = 'Checking status...'
    elif not state.is_installed:
        status = 'Not installed'
    elif not state.is_active:
        status = 'Not running'
    elif not state.is_signed_in:
        status = 'Needs authentication'
    else:
        status = 'Unknown state'
    return status


ROOT_MENU = HeadedMenu(
    title='RPi Connect',
    heading='RPi Connect',
    sub_heading=status,
    items=actions,
    placeholder='',
)


def generate_rpi_connect_menu() -> HeadedMenu:
    create_task(check_status())
    return ROOT_MENU


def init_service() -> None:
    store.dispatch(
        RegisterSettingAppAction(
            menu_item=ActionItem(
                label='RPi Connect',
                icon='',
                action=generate_rpi_connect_menu,
            ),
            category=SettingsCategory.REMOTE,
        ),
    )

    create_task(check_is_active())


Builder.load_file(
    pathlib.Path(__file__)
    .parent.joinpath('rpi_connect_qrcode_page.kv')
    .resolve()
    .as_posix(),
)



# ruff: noqa: D100
from __future__ import annotations

from typing import TYPE_CHECKING

from ubo_app.logger import logger

if TYPE_CHECKING:
    import cv2
    import numpy as np
    from numpy._typing._array_like import NDArray


class OpenCVCameraBackend:
    """Camera backend implementation using OpenCV for macOS/Linux."""

    def __init__(self, width: int, height: int, camera_index: int) -> None:
        """Initialize the OpenCV camera backend.

        Args:
            width: Desired frame width
            height: Desired frame height
            camera_index: Camera device index

        """
        self._width = width
        self._height = height
        self._camera_index = camera_index
        self._capture: cv2.VideoCapture | None = None
        self._initialize()

    def _initialize(self) -> None:
        """Initialize the OpenCV VideoCapture."""
        try:
            import cv2

            self._capture = cv2.VideoCapture(self._camera_index)

            if not self._capture.isOpened():
                logger.error('Failed to open camera at index %d', self._camera_index)
                self._capture = None
                return

            # Set camera properties
            self._capture.set(cv2.CAP_PROP_FRAME_WIDTH, self._width)
            self._capture.set(cv2.CAP_PROP_FRAME_HEIGHT, self._height)

            # Verify camera is working by capturing a test frame
            ret, _ = self._capture.read()
            if not ret:
                logger.error('Camera opened but failed to capture test frame')
                self._capture.release()
                self._capture = None
                return

            logger.info(
                'OpenCV camera initialized successfully (index: %d, resolution: %dx%d)',
                self._camera_index,
                self._width,
                self._height,
            )
        except Exception:
            logger.exception('Failed to initialize OpenCV camera')
            self._capture = None

    def start(self) -> None:
        """Start the camera and allow time for auto-exposure adjustment."""
        if self._capture is None:
            self._initialize()

        # Warm-up: Capture and discard frames to allow camera auto-exposure
        # to adjust, especially important in low-light conditions
        if self._capture:
            import time

            logger.info('Warming up camera for auto-exposure adjustment...')
            for _ in range(3):
                self._capture.read()
                time.sleep(0.1)  # 100ms between frames
            logger.info('Camera warm-up complete')

    def stop(self) -> None:
        """Stop the camera (no-op for OpenCV, use close() to release)."""

    def close(self) -> None:
        """Release camera resources."""
        if self._capture:
            self._capture.release()
            self._capture = None

    def capture_array(self, stream: str = 'main') -> NDArray[np.uint8] | None:
        """Capture a frame from the camera.

        Args:
            stream: Ignored for OpenCV (for compatibility with PiCamera2)

        Returns:
            Numpy array containing the frame data in RGB format, or None if
            capture fails

        """
        _ = stream  # Unused but required for protocol compatibility
        if not self._capture:
            return None

        import cv2
        import numpy as np

        ret, frame = self._capture.read()

        if not ret or frame is None:
            logger.warning('Failed to capture frame from camera')
            return None

        # Convert BGR (OpenCV default) to RGB (expected by the rest of the code)
        frame_rgb = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)

        # Rotate 180 degrees to correct upside-down orientation
        frame_rgb = cv2.rotate(frame_rgb, cv2.ROTATE_180)

        # Resize if needed to match expected dimensions
        if frame_rgb.shape[0] != self._height or frame_rgb.shape[1] != self._width:
            frame_rgb = cv2.resize(
                frame_rgb,
                (self._width, self._height),
                interpolation=cv2.INTER_LINEAR,
            )

        return frame_rgb.astype(np.uint8)

    def configure(self, config: dict | str) -> None:
        """Configure the camera.

        Args:
            config: Configuration dictionary (ignored for OpenCV)

        """
        # OpenCV doesn't support complex configuration like PiCamera2
        # This is a no-op for compatibility

    def set_controls(self, controls: dict) -> None:
        """Set camera controls.

        Args:
            controls: Dictionary of control parameters (ignored for OpenCV)

        """
        # OpenCV has limited control options compared to PiCamera2
        # This is a no-op for compatibility



"""Needs-Setup mixin abstract base class."""

from __future__ import annotations

import abc
from typing import final

from typing_extensions import override

from ubo_app.colors import DANGER_COLOR
from ubo_app.engines.abstraction.ai_provider_mixin import AIProviderMixin
from ubo_app.engines.abstraction.background_running_mixin import BackgroundRunningMixin
from ubo_app.engines.abstraction.engine import EngineMixin
from ubo_app.utils.async_ import create_task

ENGINE_ERROR_NOTIFICATION_ID = 'ubo:engine-error:{engine}'


class NeedsSetupMixin(EngineMixin, abc.ABC):
    """Base class for engines that require setup."""

    @override
    def __init__(self, *, label: str | None = None) -> None:
        """Initialize the NeedsSetupMixin."""
        super().__init__(label=label)
        if hasattr(self, 'run') and callable(self.run):
            self._original_run = self.run
            self.run = self._checked_run

    @property
    @abc.abstractmethod
    def not_setup_message(self) -> str:
        """The message to display when the engine is not set up."""

    @property
    @abc.abstractmethod
    def is_setup(self) -> bool:
        """Check if the engine is set up."""
        raise NotImplementedError

    @abc.abstractmethod
    async def _setup(self) -> None:
        """Perform the setup for the engine."""
        raise NotImplementedError

    @final
    def setup(self) -> None:
        """Set up the engine."""
        if isinstance(self, AIProviderMixin):
            from ubo_app.store.main import store
            from ubo_app.store.services.assistant import AssistantUpdateProvidersAction

            create_task(
                self._setup(),
                lambda task: task.add_done_callback(
                    lambda _: store.dispatch(AssistantUpdateProvidersAction()),
                ),
            )
        else:
            create_task(self._setup())

    def _checked_run(self) -> bool:
        """Check if the engine is set up before running."""
        if isinstance(self, BackgroundRunningMixin) and not self.is_setup:
            from ubo_app.store.main import store
            from ubo_app.store.services.notifications import (
                Notification,
                NotificationActionItem,
                NotificationsAddAction,
            )

            store.dispatch(
                NotificationsAddAction(
                    notification=Notification(
                        id=ENGINE_ERROR_NOTIFICATION_ID.format(engine=self.name),
                        title='Speech Recognition',
                        content=self.not_setup_message,
                        color=DANGER_COLOR,
                        actions=[
                            NotificationActionItem(
                                key='setup',
                                label='Set Up',
                                action=self.setup,
                                icon='󰒓',
                            ),
                        ],
                    ),
                ),
            )
            return False
        return self._original_run()



"""ElevenLabs engine interface."""

import re

from typing_extensions import override

from ubo_app.constants.assistant import (
    ELEVENLABS_API_KEY_PATTERN,
    ELEVENLABS_API_KEY_SECRET_ID,
    ELEVENLABS_VOICE_ID,
    ELEVENLABS_VOICE_ID_PATTERN,
)
from ubo_app.engines.abstraction.ai_provider_mixin import AIProviderMixin
from ubo_app.engines.abstraction.needs_setup_mixin import NeedsSetupMixin
from ubo_app.engines.abstraction.remote_mixin import RemoteMixin
from ubo_app.store.input.types import (
    InputFieldDescription,
    InputFieldType,
    QRCodeInputDescription,
    WebUIInputDescription,
)
from ubo_app.store.services.speech_synthesis import ReadableInformation
from ubo_app.utils import secrets
from ubo_app.utils.input import ubo_input


class ElevenLabsEngine(NeedsSetupMixin, AIProviderMixin, RemoteMixin):
    """ElevenLabs engine."""

    @property
    def name(self) -> str:
        """The internal name of the ElevenLabs engine."""
        return 'elevenlabs'

    @property
    def label(self) -> str:
        """The display label for the ElevenLabs engine."""
        return 'ElevenLabs'

    @property
    def not_setup_message(self) -> str:
        """Message shown when the ElevenLabs service API key is not set."""
        return (
            'ElevenLabs service API key and voice ID are not set. '
            'You can set them in the settings.'
        )

    @property
    @override
    def is_setup(self) -> bool:
        """Check if the ElevenLabs engine is set up."""
        api_key = secrets.read_secret(ELEVENLABS_API_KEY_SECRET_ID)
        voice_id = secrets.read_secret(ELEVENLABS_VOICE_ID)

        return (
            bool(api_key)
            and bool(voice_id)
            and re.match(ELEVENLABS_API_KEY_PATTERN, api_key) is not None
            and re.match(ELEVENLABS_VOICE_ID_PATTERN, voice_id) is not None
        )

    async def _setup(self) -> None:
        _, result = await ubo_input(
            title='ElevenLabs Configuration',
            prompt='Enter your ElevenLabs API key and voice ID.',
            descriptions=[
                WebUIInputDescription(
                    fields=[
                        InputFieldDescription(
                            name='api_key',
                            type=InputFieldType.TEXT,
                            label='API Key',
                            description='Enter your ElevenLabs API key',
                            required=True,
                            pattern=ELEVENLABS_API_KEY_PATTERN,
                        ),
                        InputFieldDescription(
                            name='voice_id',
                            type=InputFieldType.TEXT,
                            label='Voice ID',
                            description='Enter your ElevenLabs voice ID',
                            required=True,
                            pattern=ELEVENLABS_VOICE_ID_PATTERN,
                        ),
                    ],
                ),
                QRCodeInputDescription(
                    title='ElevenLabs Configuration',
                    instructions=ReadableInformation(
                        text='Convert your ElevenLabs API key and voice ID to a QR '
                        'code in the format "api_key:voice_id" and hold it in front '
                        'of the camera to scan it.',
                        picovoice_text='Convert your ElevenLabs API key and voice ID '
                        'to a {QR|K Y UW AA R} code in the format "API key colon '
                        'voice ID" and hold it in front of the camera to scan it.',
                    ),
                    pattern=(
                        r'(?P<api_key>' + ELEVENLABS_API_KEY_PATTERN + r'):'
                        r'(?P<voice_id>' + ELEVENLABS_VOICE_ID_PATTERN + r')'
                    ),
                ),
            ],
        )
        secrets.write_secret(
            key=ELEVENLABS_API_KEY_SECRET_ID,
            value=result.data['api_key'],
        )
        secrets.write_secret(
            key=ELEVENLABS_VOICE_ID,
            value=result.data['voice_id'],
        )



"""Piper engine interface."""

from __future__ import annotations

import asyncio
import hashlib
import json
import shutil
from functools import reduce

import aiostream
from typing_extensions import override
from ubo_gui.constants import WARNING_COLOR

from ubo_app.colors import DANGER_COLOR, INFO_COLOR
from ubo_app.constants.assistant import (
    PIPER_DOWNLOAD_NOTIFICATION_ID,
    PIPER_MODEL_HASH,
    PIPER_MODEL_JSON_PATH,
    PIPER_MODEL_PATH,
    PIPER_MODEL_URL,
)
from ubo_app.engines.abstraction.ai_provider_mixin import AIProviderMixin
from ubo_app.engines.abstraction.needs_setup_mixin import NeedsSetupMixin
from ubo_app.store.main import store
from ubo_app.store.services.assistant import AssistantUpdateProvidersAction
from ubo_app.store.services.notifications import (
    Chime,
    Notification,
    NotificationActionItem,
    NotificationDisplayType,
    NotificationsAddAction,
)
from ubo_app.store.services.speech_synthesis import ReadableInformation
from ubo_app.utils.async_ import create_task
from ubo_app.utils.download import download_file


class PiperEngine(NeedsSetupMixin, AIProviderMixin):
    """Piper engine."""

    @property
    def name(self) -> str:
        """The internal name of the Piper engine."""
        return 'piper'

    @property
    def label(self) -> str:
        """The display label for the Piper engine."""
        return 'Piper'

    @property
    def not_setup_message(self) -> str:
        """Message shown when the Piper service API key is not set."""
        return 'Piper model path does not exist. Please download it in the settings.'

    def _update_download_notification(self, *, progress: float) -> None:
        extra_information = ReadableInformation(
            text="""\
The download progress is shown in the radial progress bar at the top left corner of \
the screen.""",
        )
        store.dispatch(
            NotificationsAddAction(
                notification=Notification(
                    id=PIPER_DOWNLOAD_NOTIFICATION_ID,
                    title='Downloading',
                    content='Piper speech synthesis model',
                    extra_information=extra_information,
                    display_type=NotificationDisplayType.FLASH
                    if progress == 1
                    else NotificationDisplayType.STICKY,
                    flash_time=1,
                    color=INFO_COLOR,
                    icon='󰇚',
                    blink=False,
                    progress=progress,
                    show_dismiss_action=progress == 1,
                    dismiss_on_close=progress == 1,
                ),
            ),
        )

    def _handle_error(self) -> None:
        from ubo_app.store.main import store

        store.dispatch(
            NotificationsAddAction(
                notification=Notification(
                    id=PIPER_DOWNLOAD_NOTIFICATION_ID,
                    title='Piper',
                    content='Failed to download',
                    display_type=NotificationDisplayType.STICKY,
                    color=DANGER_COLOR,
                    icon='󰜺',
                    chime=Chime.FAILURE,
                ),
            ),
        )
        shutil.rmtree(PIPER_MODEL_PATH, ignore_errors=True)

    def _download_piper_model(self) -> None:
        """Download Piper model."""
        shutil.rmtree(PIPER_MODEL_PATH, ignore_errors=True)

        self._update_download_notification(progress=0)

        async def download() -> None:
            try:
                PIPER_MODEL_PATH.parent.mkdir(parents=True, exist_ok=True)

                async for download_report in aiostream.stream.ziplatest(
                    download_file(url=PIPER_MODEL_URL, path=PIPER_MODEL_PATH),
                    download_file(
                        url=f'{PIPER_MODEL_URL}.json',
                        path=PIPER_MODEL_JSON_PATH,
                    ),
                    default=(0, None),
                ):
                    downloaded_bytes, size = reduce(
                        lambda accumulator, report: (
                            report[0] + accumulator[0],
                            (report[1] or 1024**2) + accumulator[1],
                        )
                        if report
                        else accumulator,
                        download_report,
                        (0, 0),
                    )
                    if size:
                        self._update_download_notification(
                            progress=min(1.0, downloaded_bytes / size),
                        )

                self._update_download_notification(progress=1.0)
                store.dispatch(
                    AssistantUpdateProvidersAction(),
                )
            except Exception:
                self._handle_error()
                raise
            finally:
                self.event.set()

        create_task(download())

    @override
    async def _setup(self) -> None:
        if self.is_setup:
            return
        from ubo_app.store.main import store

        self.event = asyncio.Event()
        store.dispatch(
            NotificationsAddAction(
                notification=Notification(
                    title='Piper Engine Setup',
                    content='Download the Piper model.',
                    color=WARNING_COLOR,
                    actions=[
                        NotificationActionItem(
                            label='Download Model',
                            icon='󰇚',
                            action=self._download_piper_model,
                        ),
                    ],
                ),
            ),
        )
        await self.event.wait()

    @property
    @override
    def is_setup(self) -> bool:
        if not PIPER_MODEL_PATH.exists() or not PIPER_MODEL_JSON_PATH.exists():
            return False

        with PIPER_MODEL_JSON_PATH.open('r') as f:
            try:
                data = json.load(f)
            except json.JSONDecodeError:
                return False
            else:
                if data['dataset'] != 'kristin':
                    return False

        # check checksum
        with PIPER_MODEL_PATH.open('rb') as f:
            sha256_hash = hashlib.sha256()

            for chunk in iter(lambda: f.read(4096), b''):
                sha256_hash.update(chunk)

            if sha256_hash.hexdigest() != PIPER_MODEL_HASH:
                return False

        return True



# ruff: noqa: D100, D103
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ubo_handle import ReducerRegistrar, register

    from ubo_app.utils.types import Subscriptions


async def setup(register_reducer: ReducerRegistrar) -> Subscriptions:
    from reducer import reducer

    register_reducer(reducer)

    from setup import init_service

    return await init_service()


register(
    service_id='ip',
    label='IP',
    setup=setup,
)



# ruff: noqa: D100, D103
from __future__ import annotations

import asyncio
import socket
from collections import defaultdict
from contextlib import suppress
from typing import TYPE_CHECKING

import psutil
from ubo_gui.menu.types import HeadlessMenu, Item, SubMenuItem

from ubo_app.logger import logger
from ubo_app.store.core.types import RegisterSettingAppAction, SettingsCategory
from ubo_app.store.main import store
from ubo_app.store.services.ip import (
    IpNetworkInterface,
    IpSetIsConnectedAction,
    IpUpdateInterfacesAction,
)
from ubo_app.utils.async_ import create_task
from ubo_app.utils.error_handlers import report_service_error

if TYPE_CHECKING:
    from collections.abc import Sequence

    from ubo_app.utils.types import Subscriptions

PING_TIMEOUT = 3.0


@store.autorun(lambda state: state.ip.interfaces)
def get_ip_addresses(interfaces: Sequence[IpNetworkInterface]) -> list[SubMenuItem]:
    if not interfaces:
        return []
    return [
        SubMenuItem(
            label=interface.name,
            icon='󰈀'
            if interface.name.startswith('eth')
            else ''
            if interface.name.startswith('wlan')
            else '󰕇'
            if interface.name.startswith('lo')
            else '󰛳',
            sub_menu=HeadlessMenu(
                title=f'󰩟{interface.name}',
                items=[
                    Item(
                        label=ip_address,
                        icon='󰩠',
                    )
                    for ip_address in interface.ip_addresses
                ],
                placeholder='No IP addresses',
            ),
        )
        for interface in interfaces
    ]


def load_network_interfaces() -> None:
    ip_addresses_by_interface = defaultdict(list)
    for interface_name, ip_addresses in psutil.net_if_addrs().items():
        for address in ip_addresses:
            if address.family == socket.AddressFamily.AF_INET:
                ip_addresses_by_interface[interface_name].append(address.address)

    store.dispatch(
        IpUpdateInterfacesAction(
            interfaces=[
                IpNetworkInterface(name=interface_name, ip_addresses=ip_addresses)
                for interface_name, ip_addresses in ip_addresses_by_interface.items()
            ],
        ),
    )


async def monitor_interfaces(end_event: asyncio.Event) -> None:
    while not end_event.is_set():
        try:
            load_network_interfaces()
        except Exception:
            logger.exception('Failed to load network interfaces')
            report_service_error()
        finally:
            await asyncio.sleep(1)


async def monitor_connections(end_event: asyncio.Event) -> None:
    received_lines: list[tuple[bytes, float]] = []
    loop = asyncio.get_event_loop()

    async def ping_process() -> None:
        while not end_event.is_set():
            process = None
            try:
                logger.debug(
                    'Starting ping subprocess to monitor internet connectivity',
                )
                process = await asyncio.create_subprocess_exec(
                    '/usr/bin/env',
                    'ping',
                    '8.8.8.8',
                    '-s',
                    '0',
                    stdout=asyncio.subprocess.PIPE,
                    stderr=asyncio.subprocess.STDOUT,
                )
                if process.stdout:
                    async for line in process.stdout:
                        if end_event.is_set():
                            logger.debug('End event set, terminating ping process')
                            process.terminate()
                            return
                        received_lines.append((line, loop.time()))
                # If we reach here, ping process exited - wait before retry
                logger.warning(
                    'Ping process exited, restarting in 1 second...',
                    extra={'return_code': process.returncode if process else None},
                )
                await asyncio.sleep(1)
            except Exception:
                logger.exception('Error in ping process, restarting in 1 second...')
                if process:
                    with suppress(Exception):
                        process.terminate()
                await asyncio.sleep(1)

    create_task(ping_process())

    previous_state = None
    while not end_event.is_set():
        received_lines = [
            line
            for line in received_lines
            if loop.time() - line[1] < PING_TIMEOUT
            and line[0].startswith(b'8 bytes from')
        ]
        is_connected = bool(
            received_lines and received_lines[-1][0].startswith(b'8 bytes from'),
        )
        if is_connected != previous_state:
            logger.info(
                'Internet connectivity state changed',
                extra={
                    'is_connected': is_connected,
                    'received_lines_count': len(received_lines),
                },
            )
            previous_state = is_connected
        store.dispatch(IpSetIsConnectedAction(is_connected=is_connected))
        await asyncio.sleep(0.25)


IpMainMenu = SubMenuItem(
    label='IP Addresses',
    icon='󰩟',
    sub_menu=HeadlessMenu(
        title='󰩟IP Addresses',
        items=get_ip_addresses,
        placeholder='No IP addresses',
    ),
)


async def init_service() -> Subscriptions:
    store.dispatch(
        RegisterSettingAppAction(
            priority=0,
            category=SettingsCategory.NETWORK,
            menu_item=IpMainMenu,
        ),
    )

    end_event = asyncio.Event()
    create_task(monitor_connections(end_event))
    create_task(monitor_interfaces(end_event))

    return [end_event.set]



# ruff: noqa: D100, D101
from __future__ import annotations

import time
from dataclasses import field
from enum import StrEnum

from immutable import Immutable
from redux import BaseAction


class Key(StrEnum):
    BACK = 'Back'
    HOME = 'Home'
    UP = 'Up'
    DOWN = 'Down'
    L1 = 'L1'
    L2 = 'L2'
    L3 = 'L3'


class KeypadAction(BaseAction):
    key: Key
    pressed_keys: set[Key]
    held_keys: set[Key] = field(default_factory=set)
    time: float = field(default_factory=time.time)


class KeypadKeyPressAction(KeypadAction): ...


class KeypadKeyHoldAction(KeypadAction): ...


class KeypadKeyUnholdAction(KeypadAction): ...


class KeypadKeyReleaseAction(KeypadAction): ...


class KeypadState(Immutable):
    depth: int = 0
    is_consumed: bool = False



"""Setup module for initializing the File System application."""

from __future__ import annotations

import functools

from constants import SELECTOR_APPLICATION_ID
from file_application import open_path
from ubo_gui.menu.types import ActionItem

from ubo_app.store.core.types import RegisterRegularAppAction
from ubo_app.store.input.types import InputCancelAction
from ubo_app.store.main import store
from ubo_app.store.services.file_system import (
    FileSystemCopyEvent,
    FileSystemMoveEvent,
    FileSystemRemoveEvent,
    FileSystemSelectEvent,
)
from ubo_app.store.services.notifications import (
    Notification,
    NotificationActionItem,
    NotificationDisplayType,
    NotificationsAddAction,
)


def init_service() -> None:
    """Initialize the service by registering the File System application."""
    store.dispatch(
        RegisterRegularAppAction(
            menu_item=ActionItem(
                label='File System',
                icon='󰉋',
                action=open_path,
            ),
            key='file-system',
        ),
    )

    def handle_open_path_event(event: FileSystemSelectEvent) -> None:
        """Open the file system path selector."""
        store.dispatch(
            NotificationsAddAction(
                notification=Notification(
                    id=SELECTOR_APPLICATION_ID.format(id=event.description.id),
                    title=event.description.title or 'Select Path',
                    content=event.description.prompt or 'Please select a path to open.',
                    icon='󰉋',
                    display_type=NotificationDisplayType.STICKY,
                    show_dismiss_action=False,
                    on_close=functools.partial(
                        store.dispatch,
                        InputCancelAction(id=event.description.id),
                    ),
                    actions=[
                        NotificationActionItem(
                            key='open-path',
                            label='Open Path Selector',
                            icon='󰉋',
                            close_notification=False,
                            action=functools.partial(
                                open_path,
                                config=event.description.selector_config,
                            ),
                        ),
                    ],
                ),
            ),
        )

    def handle_copy_event(event: FileSystemCopyEvent) -> None:
        from shutil import copyfile, copytree

        for source in event.sources:
            if source.is_dir():
                copytree(source, event.destination / source.name)
            else:
                copyfile(source, event.destination / source.name)

    def handle_move_event(event: FileSystemMoveEvent) -> None:
        from shutil import move

        for source in event.sources:
            move(source, event.destination / source.name)

    def handle_remove_event(event: FileSystemRemoveEvent) -> None:
        from shutil import rmtree

        for source in event.paths:
            if source.is_dir():
                rmtree(source)
            else:
                source.unlink(missing_ok=True)

    store.subscribe_event(FileSystemSelectEvent, handle_open_path_event)
    store.subscribe_event(FileSystemCopyEvent, handle_copy_event)
    store.subscribe_event(FileSystemMoveEvent, handle_move_event)
    store.subscribe_event(FileSystemRemoveEvent, handle_remove_event)



"""Utility functions for camera service."""

from __future__ import annotations

from ubo_app.logger import logger


def detect_available_cameras_picamera2() -> list[int]:
    """Detect available Picamera2 cameras on Raspberry Pi.

    Returns:
        List of working camera indices

    """
    try:
        from picamera2.picamera2 import Picamera2

        available = []
        logger.info('Detecting available Picamera2 cameras...')

        # Get list of available cameras
        cameras = Picamera2.global_camera_info()

        for i, camera_info in enumerate(cameras):
            logger.info(
                'Found camera at index {index}: {info}',
                extra={'index': i, 'info': camera_info},
            )
            available.append(i)

        logger.info(
            'Picamera2 detection complete: {count} camera(s) found',
            extra={'count': len(available), 'indices': available},
        )
    except (ImportError, RuntimeError, OSError) as e:
        logger.exception(
            'Failed to detect Picamera2 cameras',
            extra={'error': e},
        )
    else:
        return available

    return []


def detect_available_cameras(max_index: int = 10) -> list[int]:
    """Detect available camera indices by attempting to open them.

    Args:
        max_index: Maximum camera index to check (default: 10)

    Returns:
        List of working camera indices

    """
    import cv2

    available = []
    logger.info('Detecting available cameras...')

    for i in range(max_index):
        cap = None
        try:
            cap = cv2.VideoCapture(i)
            if cap.isOpened():
                # Try to read a frame to verify camera actually works
                ret, frame = cap.read()
                if ret and frame is not None:
                    available.append(i)
                    logger.info(
                        'Found working camera at index {index}',
                        extra={'index': i},
                    )
        except (cv2.error, OSError, RuntimeError) as e:
            logger.debug(
                'Failed to open camera at index {index}',
                extra={'index': i, 'error': e},
            )
        finally:
            if cap is not None:
                cap.release()

    logger.info(
        'Camera detection complete: {count} camera(s) found',
        extra={'count': len(available), 'indices': available},
    )
    return available



# ruff: noqa: D100, D103
import platform

from ubo_app.constants import CACHE_PATH, DATA_PATH


def get_cli_tool_name() -> str:  # noqa: C901
    # Determine the OS type and architecture
    os_type = platform.system().lower()
    arch = platform.machine().lower()

    if os_type == 'linux':
        if 'aarch64' in arch or 'arm64' in arch:
            return 'cli-alpine-arm64'
        if 'arm' in arch:
            return 'cli-alpine-armhf'
        if 'x86_64' in arch:
            return 'cli-alpine-x64'

    if os_type == 'darwin':
        if 'arm64' in arch:
            return 'cli-darwin-arm64'
        if 'x86_64' in arch:
            return 'cli-darwin-x64'

    if os_type == 'windows':
        if 'arm64' in arch:
            return 'cli-win32-arm64'
        if 'x86_64' in arch:
            return 'cli-win32-x64'
        if 'x86' in arch:
            return 'cli-win32'

    raise ValueError


CODE_BINARY_URL = (
    f'https://code.visualstudio.com/sha/download?build=stable&os={get_cli_tool_name()}'
)

CODE_BINARY_PATH = DATA_PATH / 'code'
CODE_DOWNLOAD_PATH = CACHE_PATH / 'code.tar.gz'



# ruff: noqa: D100, D103
from __future__ import annotations

from dataclasses import replace
from typing import TYPE_CHECKING

from redux import CompleteReducerResult, InitAction, InitializationActionError

from ubo_app.store.input.types import (
    InputAction,
    InputCancelAction,
    InputDemandAction,
    InputResolveAction,
    WebUIInputDescription,
)
from ubo_app.store.services.notifications import (
    NotificationsAction,
    NotificationsClearByIdAction,
)
from ubo_app.store.services.web_ui import (
    WebUIInitializeEvent,
    WebUIState,
    WebUIStopEvent,
)

if TYPE_CHECKING:
    from redux import ReducerResult

DispatchAction = InputCancelAction | NotificationsAction


def reducer(
    state: WebUIState | None,
    action: InputAction,
) -> ReducerResult[
    WebUIState,
    DispatchAction,
    WebUIInitializeEvent | WebUIStopEvent,
]:
    if state is None:
        if isinstance(action, InitAction):
            return WebUIState(active_inputs=[])
        raise InitializationActionError(action)

    match action:
        case InputDemandAction(description=WebUIInputDescription() as description):
            return CompleteReducerResult(
                state=replace(
                    state,
                    active_inputs=[*state.active_inputs, description],
                ),
                events=[WebUIInitializeEvent(description=description)],
            )

        case InputResolveAction(id=id):
            new_active_inputs = [
                description
                for description in state.active_inputs
                if description.id != id
            ]
            should_dispatch_stop_event = (
                len(state.active_inputs) > 0 and len(new_active_inputs) == 0
            )
            return CompleteReducerResult(
                state=replace(
                    state,
                    active_inputs=new_active_inputs,
                ),
                actions=[NotificationsClearByIdAction(id=f'web_ui:pending:{id}')],
                events=[WebUIStopEvent()] if should_dispatch_stop_event else [],
            )

        case _:
            return state



# ruff: noqa: D100, D101
from __future__ import annotations

import socket
from dataclasses import field
from enum import StrEnum

from immutable import Immutable
from redux import BaseAction, BaseEvent

from ubo_app.utils.dataclass import default_provider
from ubo_app.utils.persistent_store import read_from_persistent_store


class SpeechSynthesisAction(BaseAction): ...


class SpeechSynthesisEvent(BaseEvent): ...


class SpeechSynthesisUpdateAccessKeyStatus(SpeechSynthesisAction):
    is_access_key_set: bool


class SpeechSynthesisEngineName(StrEnum):
    PIPER = 'piper'
    PICOVOICE = 'picovoice'


class SpeechSynthesisSetSelectedEngineAction(SpeechSynthesisAction):
    engine_name: SpeechSynthesisEngineName


class ReadableInformation(Immutable):
    text: str
    piper_text: str = field(
        default_factory=default_provider(
            ['text'],
            lambda text: text.replace('{{hostname}}', f'{socket.gethostname()}.local'),
        ),
    )
    picovoice_text: str = field(
        default_factory=default_provider(
            ['text'],
            lambda text: text.replace('{{hostname}}', f'{socket.gethostname()}.local'),
        ),
    )

    def __post_init__(self) -> None:
        """Replace `{{hostname}}` with the current hostname."""
        object.__setattr__(
            self,
            'text',
            self.text.replace(
                '{{hostname}}',
                f'{socket.gethostname()}.local',
            ),
        )

    def __add__(
        self,
        other: ReadableInformation,
    ) -> ReadableInformation:
        """Concatenate two `ReadableInformation` objects."""
        return ReadableInformation(
            text=self.text + other.text,
            piper_text=self.piper_text + other.piper_text,
            picovoice_text=self.picovoice_text + other.picovoice_text,
        )


class SpeechSynthesisReadTextAction(SpeechSynthesisAction):
    information: ReadableInformation
    speech_rate: float | None = None
    engine: SpeechSynthesisEngineName | None = None


class SpeechSynthesisSynthesizeTextEvent(SpeechSynthesisEvent):
    information: ReadableInformation
    speech_rate: float | None = None


class SpeechSynthesisState(Immutable):
    is_access_key_set: bool | None = None
    selected_engine: SpeechSynthesisEngineName = field(
        default=read_from_persistent_store(
            key='speech_synthesis:selected_engine',
            default=SpeechSynthesisEngineName.PIPER,
        ),
    )



# ruff: noqa: T201
"""Generate proto files from actions, events and states defined in Python files."""

from __future__ import annotations

import ast
import importlib
import sys
from pathlib import Path
from typing import TYPE_CHECKING

import betterproto.casing
from ubo_app.rpc.generator._module_transformer import _ProtoGenerator

if TYPE_CHECKING:
    from types import ModuleType


def _generate_operations_proto(
    output_directory: Path,
    actions: list,
    events: list,
) -> None:
    proto = '\n'
    if actions:
        proto += 'message Action {\n'
        proto += '  oneof action {\n'
        for i, (action, _) in enumerate(sorted(generator.actions), 1):
            # proto += f"""    {package_name}.v1.{action} {
            proto += (
                f"""    {action} {betterproto.casing.snake_case(action)} = {i};\n"""
            )
        proto += '  }\n'
        proto += '}\n\n'
    if events:
        proto += 'message Event {\n'
        proto += '  oneof event {\n'
        for i, (event, _) in enumerate(sorted(generator.events), 1):
            # proto += f"""    {package_name}.v1.{event} {
            proto += f"""    {event} {betterproto.casing.snake_case(event)} = {i};\n"""
        proto += '  }\n'
        proto += '}\n\n'
    operations_proto_path = output_directory / 'ubo' / 'v1' / 'ubo.proto'
    with operations_proto_path.open('a') as file:
        file.write(proto)


def parse(input_module: ModuleType, actions: list, events: list) -> _ProtoGenerator:
    """Generate proto files from actions, events and states defined in Python files."""
    if not input_module.__file__:
        msg = 'Module must be a file'
        raise ValueError(msg)
    with Path(input_module.__file__).open() as file:
        tree = ast.parse(file.read())

    generator = _ProtoGenerator(module=input_module, actions=actions, events=events)
    generator.visit(tree)

    return generator


if __name__ == '__main__':
    print('🚀 Generating proto files...')
    output_directory = Path('ubo_app/rpc/proto/')
    import ubo_app.utils.persistent_store

    ubo_app.utils.persistent_store.read_from_persistent_store = lambda *_, **__: ''

    import ubo_app.store.core.types
    import ubo_app.store.input.types
    import ubo_app.store.settings.types
    import ubo_app.store.status_icons.types
    import ubo_app.store.ubo_actions
    import ubo_app.store.update_manager.types
    import ubo_gui.menu.types

    generators: list[_ProtoGenerator] = []

    actions = []
    events = []

    generators.append(parse(ubo_gui.menu.types, actions, events))
    generators.append(parse(ubo_app.store.core.types, actions, events))
    generators.append(
        parse(ubo_app.store.ubo_actions, actions, events),
    )
    generators.append(parse(ubo_app.store.input.types, actions, events))
    generators.append(parse(ubo_app.store.settings.types, actions, events))
    generators.append(parse(ubo_app.store.status_icons.types, actions, events))
    generators.append(parse(ubo_app.store.update_manager.types, actions, events))
    generators.extend(
        parse(
            importlib.import_module(f'ubo_app.store.services.{file.stem}'),
            actions,
            events,
        )
        for file in sorted(Path('ubo_app/store/services/').glob('*.py'))
    )

    (output_directory / 'ubo' / 'v1').mkdir(
        exist_ok=True,
        parents=True,
    )
    with (output_directory / 'ubo' / 'v1' / 'ubo.proto').open('w') as file:
        file.write('syntax = "proto3";\n\n')
        file.write('package ubo.v1;\n\n')
        file.write('import "package_info/v1/package_info.proto";\n\n')
        for generator in generators:
            sys.stdout.write(f'⚡ Generating proto for {generator.package_name} .')

            file.write(generator.generate_proto())

            print(' Done')

    """ This version is for the case where we have a separate proto file for each
    package, it is currently not possible because proto files do not support import
    loops, and Action needs NotificationsAddAction which needs NotificationActionItem
    which needs Action

    for generator in generators:
        proto_definitions, dependencies = generator.generate_proto()

        (output_directory / generator.package_name / 'v1').mkdir(
            exist_ok=True,
            parents=True,
        )
        with (
            output_directory
            / generator.package_name
            / 'v1'
            / f'{generator.package_name}.proto'
        ).open('w') as file:
            file.write('syntax = "proto3";\n\n')
            file.write(f'package {generator.package_name}.v1;\n\n')
            if dependencies:
                for dependency in dependencies:
                    file.write(f'import "{dependency}/v1/{dependency}.proto";\n')
                file.write('\n')
            file.write(proto_definitions)
    """

    _generate_operations_proto(output_directory, actions, events)

    print('🎉 Proto files generated successfully!')



"""Renders the QR code of the container's url (ip and port)."""

from __future__ import annotations

import pathlib

from kivy.lang.builder import Builder
from kivy.properties import ListProperty, NumericProperty, StringProperty

from ubo_app.utils.gui import UboPageWidget


class DockerQRCodePage(UboPageWidget):
    """QR code for the container's url (ip and port)."""

    ips: list[str] = ListProperty()
    port: str = StringProperty()
    index: int = NumericProperty(0)

    def go_down(self: DockerQRCodePage) -> None:
        """Go down."""
        self.index = (self.index + 1) % len(self.ips)
        self.ids.slider.animated_value = len(self.ips) - 1 - self.index

    def go_up(self: DockerQRCodePage) -> None:
        """Go up."""
        self.index = (self.index - 1) % len(self.ips)
        self.ids.slider.animated_value = len(self.ips) - 1 - self.index


Builder.load_file(
    pathlib.Path(__file__)
    .parent.joinpath('docker_qrcode_page.kv')
    .resolve()
    .as_posix(),
)



"""MCP server storage and management utilities."""

from __future__ import annotations

import json
import logging
import shutil
import uuid
from typing import TYPE_CHECKING

from ubo_app.constants.assistant import ASSISTANT_MCP_SERVERS_PATH
from ubo_app.store.services.assistant import McpServerMetadata, McpServerType

if TYPE_CHECKING:
    from collections.abc import Sequence
    from pathlib import Path

logger = logging.getLogger(__name__)


def load_mcp_servers() -> dict[str, McpServerMetadata]:
    """Load all MCP servers from filesystem.

    Returns:
        Dictionary mapping server_id to McpServerMetadata

    """
    servers: dict[str, McpServerMetadata] = {}

    if not ASSISTANT_MCP_SERVERS_PATH.exists():
        ASSISTANT_MCP_SERVERS_PATH.mkdir(parents=True, exist_ok=True)
        return servers

    for server_dir in ASSISTANT_MCP_SERVERS_PATH.iterdir():
        if not server_dir.is_dir():
            continue

        config_file = server_dir / 'config.json'
        if not config_file.exists():
            logger.warning(
                'MCP server directory missing config.json',
                extra={'server_dir': server_dir.name},
            )
            continue

        try:
            with config_file.open() as f:
                data = json.load(f)

            server_type = McpServerType(data['type'])
            config = data['config']

            # Extract server name from directory name (format: {name}_{uuid})
            server_id = server_dir.name
            name_parts = server_id.rsplit('_', 1)
            name = name_parts[0] if len(name_parts) == 2 else server_id  # noqa: PLR2004

            servers[server_id] = McpServerMetadata(
                server_id=server_id,
                name=name,
                type=server_type,
                config=config,
            )

        except (json.JSONDecodeError, KeyError, ValueError) as e:
            logger.exception(
                'Failed to load MCP server config',
                extra={'server_dir': server_dir.name, 'error': str(e)},
            )

    return servers


def save_mcp_server(
    name: str,
    server_type: McpServerType,
    config: str,
) -> str:
    """Save MCP server configuration to filesystem.

    Args:
        name: User-friendly server name
        server_type: Type of MCP server (stdio or sse)
        config: Server configuration (JSON string for stdio, URL string for sse)

    Returns:
        server_id: The generated server ID

    """
    # Generate server_id: {name}_{short_uuid}
    short_uuid = uuid.uuid4().hex[:8]
    server_id = f'{name}_{short_uuid}'

    server_dir = ASSISTANT_MCP_SERVERS_PATH / server_id
    server_dir.mkdir(parents=True, exist_ok=True)

    config_file = server_dir / 'config.json'
    data = {
        'type': server_type.value,
        'config': config,
        'enabled': True,  # Default to enabled
    }

    with config_file.open('w') as f:
        json.dump(data, f, indent=2)

    logger.info(
        'Saved MCP server configuration',
        extra={'server_id': server_id, 'server_name': name, 'type': server_type.value},
    )

    return server_id


def toggle_mcp_server(server_id: str) -> bool:
    """Toggle MCP server enabled state.

    Args:
        server_id: The server ID to toggle

    Returns:
        New enabled state (True if enabled, False if disabled)

    """
    server_dir = ASSISTANT_MCP_SERVERS_PATH / server_id
    config_file = server_dir / 'config.json'

    if not config_file.exists():
        logger.warning(
            'Attempted to toggle non-existent MCP server',
            extra={'server_id': server_id},
        )
        return False

    try:
        with config_file.open() as f:
            data = json.load(f)

        # Toggle the enabled state (default to False if not present)
        current_state = data.get('enabled', False)
        new_state = not current_state
        data['enabled'] = new_state

        with config_file.open('w') as f:
            json.dump(data, f, indent=2)

        logger.info(
            'Toggled MCP server state',
            extra={'server_id': server_id, 'enabled': new_state},
        )
    except (json.JSONDecodeError, OSError) as e:
        logger.exception(
            'Failed to toggle MCP server state',
            extra={'server_id': server_id, 'error': str(e)},
        )
        return False
    else:
        return new_state


def delete_mcp_server(server_id: str) -> None:
    """Delete MCP server from filesystem.

    Args:
        server_id: The server ID to delete

    """
    server_dir = ASSISTANT_MCP_SERVERS_PATH / server_id

    if not server_dir.exists():
        logger.warning(
            'Attempted to delete non-existent MCP server',
            extra={'server_id': server_id},
        )
        return

    shutil.rmtree(server_dir)
    logger.info('Deleted MCP server', extra={'server_id': server_id})


def validate_stdio_config(config_str: str) -> tuple[bool, str, dict | None]:
    """Validate stdio MCP server configuration JSON.

    Args:
        config_str: JSON string to validate

    Returns:
        Tuple of (is_valid, error_message, parsed_config)

    """
    try:
        config = json.loads(config_str)
    except json.JSONDecodeError as e:
        return False, f'Invalid JSON: {e}', None

    # Check for mcpServers key
    if 'mcpServers' not in config:
        return False, 'Missing "mcpServers" key in configuration', None

    mcp_servers = config['mcpServers']
    if not isinstance(mcp_servers, dict):
        return False, '"mcpServers" must be an object', None

    if not mcp_servers:
        return False, '"mcpServers" cannot be empty', None

    # Check that there's exactly one server
    if len(mcp_servers) != 1:
        return (
            False,
            'Configuration must contain exactly one server per form',
            None,
        )

    # Validate server structure
    server_name, server_config = next(iter(mcp_servers.items()))

    if not isinstance(server_config, dict):
        return False, f'Server "{server_name}" configuration must be an object', None

    # Check required fields for stdio
    if 'command' not in server_config:
        return False, f'Server "{server_name}" missing required "command" field', None

    return True, '', config


def validate_sse_url(url: str) -> tuple[bool, str]:
    """Validate SSE URL format.

    Args:
        url: URL string to validate

    Returns:
        Tuple of (is_valid, error_message)

    """
    if not url or not url.strip():
        return False, 'URL cannot be empty'

    url = url.strip()

    # Basic URL validation
    if not url.startswith(('http://', 'https://')):
        return False, 'URL must start with http:// or https://'

    # Check for basic URL structure
    if len(url) < 10:  # noqa: PLR2004
        return False, 'URL appears to be too short'

    return True, ''


def get_server_directories() -> Sequence[Path]:
    """Get all MCP server directories.

    Returns:
        List of Path objects for server directories

    """
    if not ASSISTANT_MCP_SERVERS_PATH.exists():
        return []

    return [
        d for d in ASSISTANT_MCP_SERVERS_PATH.iterdir() if d.is_dir()
    ]



# ruff: noqa: D100, D103
from ubo_app.store.main import store
from ubo_app.store.services.rgb_ring import RgbRingCommandEvent, RgbRingPulseAction
from ubo_app.utils.eeprom import get_eeprom_data
from ubo_app.utils.server import send_command


def init_service() -> None:
    eeprom_data = get_eeprom_data()

    if (
        (led := eeprom_data.get('led')) is None
        or led.get('model') != 'neopixel'
    ):
        return

    async def handle_rgb_ring_command(event: RgbRingCommandEvent) -> None:
        await send_command('led', *event.command)

    store.subscribe_event(RgbRingCommandEvent, handle_rgb_ring_command)

    store.dispatch(RgbRingPulseAction(repetitions=2, wait=180))



# ruff: noqa: D100, D103

from __future__ import annotations

from dataclasses import replace

from constants import SELECTOR_APPLICATION_ID
from redux import (
    CompleteReducerResult,
    InitAction,
    InitializationActionError,
    ReducerResult,
)

from ubo_app.store.input.types import (
    InputAction,
    InputDemandAction,
    InputMethod,
    InputProvideAction,
    InputResolveAction,
    InputResult,
    PathInputDescription,
)
from ubo_app.store.services.file_system import (
    FileSystemAction,
    FileSystemCopyAction,
    FileSystemCopyEvent,
    FileSystemEvent,
    FileSystemMoveAction,
    FileSystemMoveEvent,
    FileSystemRemoveAction,
    FileSystemRemoveEvent,
    FileSystemReportSelectionAction,
    FileSystemSelectEvent,
    FileSystemState,
)
from ubo_app.store.services.notifications import NotificationsClearByIdAction

DispatchAction = NotificationsClearByIdAction | InputProvideAction


def pop_queue(
    state: FileSystemState,
    *,
    actions: list[DispatchAction] | None = None,
    events: list[FileSystemEvent] | None = None,
) -> ReducerResult[
    FileSystemState,
    DispatchAction,
    FileSystemEvent,
]:
    actions = actions or []
    events = events or []

    actions += [
        NotificationsClearByIdAction(
            id=SELECTOR_APPLICATION_ID.format(id=state.queue[0].id),
        ),
    ]

    _, *queue = state.queue
    if queue:
        events.append(FileSystemSelectEvent(description=queue[0]))
    return CompleteReducerResult(
        state=replace(state, queue=queue),
        actions=actions,
        events=events,
    )


def reducer(
    state: FileSystemState | None,
    action: FileSystemAction | InputAction,
) -> ReducerResult[
    FileSystemState,
    DispatchAction,
    FileSystemEvent,
]:
    if state is None:
        if isinstance(action, InitAction):
            return FileSystemState(queue=[])

        raise InitializationActionError(action)

    match action:
        case FileSystemCopyAction():
            return CompleteReducerResult(
                state=state,
                events=[
                    FileSystemCopyEvent(
                        sources=action.sources,
                        destination=action.destination,
                    ),
                ],
            )

        case FileSystemMoveAction():
            return CompleteReducerResult(
                state=state,
                events=[
                    FileSystemMoveEvent(
                        sources=action.sources,
                        destination=action.destination,
                    ),
                ],
            )

        case FileSystemRemoveAction():
            return CompleteReducerResult(
                state=state,
                events=[FileSystemRemoveEvent(paths=action.paths)],
            )

        case InputDemandAction() if isinstance(
            action.description,
            PathInputDescription,
        ):
            return CompleteReducerResult(
                state=replace(state, queue=[*state.queue, action.description]),
                events=[]
                if state.queue
                else [FileSystemSelectEvent(description=action.description)],
            )

        case InputResolveAction():
            if state.queue and state.queue[0].id == action.id:
                return pop_queue(state)
            return replace(
                state,
                queue=[
                    description
                    for description in state.queue
                    if description.id != action.id
                ],
            )

        case FileSystemReportSelectionAction():
            return CompleteReducerResult(
                state=state,
                actions=[
                    InputProvideAction(
                        id=state.queue[0].id,
                        value=action.path.as_posix(),
                        result=InputResult(
                            data={'path': action.path.as_posix()},
                            files={},
                            method=InputMethod.PATH_SELECTOR,
                        ),
                    ),
                ],
            )

        case _:
            return state



"""Update manager module."""

from __future__ import annotations

import asyncio
import functools
import importlib.metadata
import shutil
import tarfile
from dataclasses import replace
from pathlib import Path
from typing import TYPE_CHECKING

import aiohttp
import requests
from debouncer import DebounceOptions, debounce
from redux import FinishAction
from ubo_gui.menu.types import HeadedMenu, HeadlessMenu, Item, SubMenuItem

from ubo_app.colors import DANGER_COLOR, INFO_COLOR, SUCCESS_COLOR
from ubo_app.constants import (
    INSTALLATION_PATH,
    PACKAGE_NAME,
    UPDATE_ASSETS_PATH,
)
from ubo_app.logger import logger
from ubo_app.store.main import store
from ubo_app.store.services.notifications import (
    Chime,
    Notification,
    NotificationActionItem,
    NotificationDispatchItem,
    NotificationDisplayType,
    NotificationsAddAction,
)
from ubo_app.store.services.speech_synthesis import ReadableInformation
from ubo_app.store.ubo_actions import UboDispatchItem
from ubo_app.store.update_manager.installed_versions import get_installed_versions
from ubo_app.store.update_manager.types import (
    UPDATE_MANAGER_NOTIFICATION_ID,
    UpdateManagerReportFailedCheckAction,
    UpdateManagerRequestCheckAction,
    UpdateManagerRequestUpdateAction,
    UpdateManagerSetVersionsAction,
    UpdateManagerState,
    UpdateManagerUpdateEvent,
    UpdateStatus,
)
from ubo_app.utils import IS_RPI
from ubo_app.utils.download import download_file
from ubo_app.utils.gui import SELECTED_ITEM_PARAMETERS, UNSELECTED_ITEM_PARAMETERS
from ubo_app.utils.server import send_command

if TYPE_CHECKING:
    from collections.abc import Callable

CURRENT_VERSION = importlib.metadata.version(PACKAGE_NAME)
if IS_RPI:
    try:
        BASE_IMAGE = Path('/etc/ubo_base_image').read_text()
    except FileNotFoundError:
        BASE_IMAGE = '[unknown]'
else:
    BASE_IMAGE = '[unknown]'
BASE_IMAGE_VARIANT = (
    (BASE_IMAGE == '[unknown]' and '[unknown]')
    or (BASE_IMAGE.endswith('-lite') and 'lite')
    or (BASE_IMAGE.endswith('-full') and 'full')
    or 'desktop'
)


@store.with_state(lambda state: state.settings.beta_versions)
@debounce(6, options=DebounceOptions(leading=True, trailing=True, time_window=6))
async def check_version(beta_versions: bool) -> None:  # noqa: FBT001
    """Check for updates."""
    logger.info('Checking for updates...', extra={'beta_versions': beta_versions})

    try:
        async with (
            aiohttp.ClientSession() as session,
            session.get(
                'https://pypi.org/pypi/ubo-app/json',
                timeout=aiohttp.ClientTimeout(total=5),
            ) as response,
        ):
            if response.status != requests.codes.ok:
                logger.error('Failed to check for updates')
                return
            data = await response.json()
            if beta_versions:
                recent_versions = [
                    version
                    for version, data in data['releases'].items()
                    if not data[0]['yanked']
                ]
                recent_versions = sorted(
                    recent_versions,
                    key=lambda x: data['releases'][x][0]['upload_time'],
                    reverse=True,
                )[:12]
            else:
                recent_versions = [
                    version
                    for version, data in data['releases'].items()
                    if '.dev' not in version and not data[0]['yanked']
                ]
                recent_versions = sorted(
                    recent_versions,
                    key=lambda x: data['releases'][x][0]['upload_time'],
                    reverse=True,
                )[:3]

            if beta_versions:
                latest_version = recent_versions[0]
            else:
                latest_version = data['info']['version']

            store.dispatch(
                with_state=lambda state: UpdateManagerSetVersionsAction(
                    flash_notification=state is None
                    or state.main.path[:2] != ['main', 'about'],
                    current_version=CURRENT_VERSION,
                    base_image_variant=BASE_IMAGE_VARIANT,
                    latest_version=latest_version,
                    recent_versions=recent_versions,
                ),
            )
    except Exception:
        logger.exception('Failed to check for updates')
        store.dispatch(UpdateManagerReportFailedCheckAction())
        return


UPDATE_PROGRESS_NOTIFICATION = Notification(
    id=UPDATE_MANAGER_NOTIFICATION_ID,
    title='Update in progress',
    content='Downloading the latest version of the install script...',
    extra_information=ReadableInformation(
        text="""\
The download progress is shown in the radial progress bar at the top left corner of \
the screen.
Once the download is complete, ubo app will quit and the new version will run.""",
    ),
    display_type=NotificationDisplayType.STICKY,
    color=INFO_COLOR,
    icon='󰇚',
    blink=False,
    progress=0,
    show_dismiss_action=False,
)


async def _update(target_version: str | None) -> None:
    store.dispatch(NotificationsAddAction(notification=UPDATE_PROGRESS_NOTIFICATION))

    shutil.rmtree(UPDATE_ASSETS_PATH, ignore_errors=True)
    UPDATE_ASSETS_PATH.mkdir(parents=True, exist_ok=True)

    packages_count_path = f'{INSTALLATION_PATH}/.packages-count'
    try:
        packages_count = int(Path(packages_count_path).read_text(encoding='utf-8'))
        _ = 1 / packages_count
    except (FileNotFoundError, ValueError, ZeroDivisionError):
        packages_count = 112
    counter = 0

    async for report in download_file(
        url=f'https://files.pythonhosted.org/packages/source/u/ubo_app/ubo_app-{target_version}.tar.gz',
        path=UPDATE_ASSETS_PATH / 'package.tar.gz',
    ):
        store.dispatch(
            NotificationsAddAction(
                notification=replace(
                    UPDATE_PROGRESS_NOTIFICATION,
                    progress=(report[0] / report[1]) * 0.05 if report[1] else 0,
                ),
            ),
        )

    with tarfile.open(
        UPDATE_ASSETS_PATH / 'package.tar.gz',
        'r:gz',
    ) as tar:
        install_file_info = next(
            (
                m
                for m in tar.getmembers()
                if m.name.endswith('ubo_app/system/scripts/install.sh')
            ),
            None,
        )
        if not install_file_info:
            msg = 'Failed to extract install.sh'
            raise RuntimeError(msg)
        install_file_info.name = 'install.sh'
        tar.extract(install_file_info, path=UPDATE_ASSETS_PATH)

    (UPDATE_ASSETS_PATH / 'install.sh').chmod(0o755)

    store.dispatch(
        NotificationsAddAction(
            notification=replace(
                UPDATE_PROGRESS_NOTIFICATION,
                content='Running the `install.sh` script...',
                progress=0.06,
            ),
        ),
    )

    def handle_package_collection_progress(line: str) -> tuple[float, str]:
        """Handle package collection progress."""
        nonlocal counter
        counter += 1
        package_name = line.partition(' ')[2].strip()
        return (
            min((counter + 1) / packages_count, 1) * 0.3 + 0.2,
            f'Downloading {package_name}',
        )

    progress_map: dict[str, tuple[float, str] | Callable[[str], tuple[float, str]]] = {
        'Installing dependencies...': (0.07, 'Installing system dependencies...'),
        'Setting up Python virtual environment...': (
            0.19,
            'Setting up Python virtual environment...',
        ),
        'Installing packages...': (0.2, 'Installing Python packages...'),
        'Collecting ': handle_package_collection_progress,
        'Installing collected packages: ': (
            0.51,
            'Installing collected packages...',
        ),
        'ubo-app installed successfully in ': (
            0.7,
            'ubo-app package installed successfully.',
        ),
        'Installing WM8960 driver...': (0.71, 'Installing WM8960 driver...'),
        'WM8960 driver installed successfully.': (
            0.8,
            'WM8960 driver installed successfully.',
        ),
        'Installing Docker...': (0.81, 'Installing Docker...'),
        'Docker installed successfully.': (
            0.9,
            'Docker installed successfully.',
        ),
        'Bootstrapping ubo-app...': (0.9, 'Bootstrapping ubo-app...'),
        'Bootstrapping completed': (1, 'Bootstrapping completed'),
    }

    async for line in await send_command(
        'update',
        target_version or '',
        has_output_stream=True,
    ):
        for key, report_ in progress_map.items():
            if line.startswith(key):
                report = report_(line) if callable(report_) else report_
                progress, message = report
                store.dispatch(
                    NotificationsAddAction(
                        notification=replace(
                            UPDATE_PROGRESS_NOTIFICATION,
                            content=message,
                            progress=progress,
                        ),
                    ),
                )
                if progress == 1:
                    await asyncio.sleep(1)
                    break
        else:
            continue
        break
    else:
        msg = 'Failed to update: install script failed, check system manager logs.'
        logger.info(msg)
        raise RuntimeError(msg)

    store.dispatch(
        NotificationsAddAction(
            notification=Notification(
                id=UPDATE_MANAGER_NOTIFICATION_ID,
                title='Update Complete',
                content="""\
Ubo App will restart now.""",
                extra_information=ReadableInformation(
                    text='In a few seconds, the ubo app should restart itself and run '
                    'the newly installed version.',
                ),
                display_type=NotificationDisplayType.STICKY,
                color=SUCCESS_COLOR,
                icon='󰄬',
                progress=1,
                show_dismiss_action=False,
            ),
        ),
    )


async def update(event_request_event: UpdateManagerUpdateEvent) -> None:
    """Update the Ubo app."""
    logger.info('Updating Ubo app...')
    target_version = event_request_event.version

    if target_version is None:
        return

    try:
        await _update(target_version)
    except Exception:
        logger.exception('Failed to update')
        store.dispatch(
            NotificationsAddAction(
                notification=Notification(
                    title='Failed to update',
                    content='Failed to update',
                    display_type=NotificationDisplayType.FLASH,
                    color=DANGER_COLOR,
                    icon='󰜺',
                    chime=Chime.FAILURE,
                ),
            ),
            UpdateManagerRequestCheckAction(),
        )
        return


def activate_version(version: Path) -> None:
    """Activate the selected version."""
    logger.info('Activating version %s', version.name)
    if not version.is_dir():
        return

    env_path = Path(INSTALLATION_PATH) / 'env'
    env_path.unlink(missing_ok=True)
    env_path.symlink_to(version, target_is_directory=True)

    store.dispatch(FinishAction())


def open_about_menu() -> HeadedMenu:
    """Get the about menu items."""
    store.dispatch(UpdateManagerRequestCheckAction())

    return HeadedMenu(
        title='About',
        heading=f'Ubo v{CURRENT_VERSION}',
        sub_heading=f'Base image: {BASE_IMAGE[:11]}\n{BASE_IMAGE[11:]}',
        items=about_menu_items,
    )


@store.autorun(lambda state: (state.update_manager, state.settings.beta_versions))
def about_menu_items(data: tuple[UpdateManagerState, bool]) -> list[Item]:
    """Get the update menu items."""
    state, beta_versions = data
    items: list[Item] = []

    recent_versions_item = SubMenuItem(
        key='recent_versions',
        label='Recent versions',
        icon='󰜉',
        sub_menu=HeadlessMenu(
            title='󰜉Recent versions',
            items=[
                UboDispatchItem(
                    label=version,
                    icon='󰜉',
                    store_action=NotificationsAddAction(
                        notification=Notification(
                            id=f'update-manager-install-{version}',
                            title=f'Install {version} now?',
                            content='Press 󰜉 button to start installation of version '
                            f'"{version}"',
                            icon='󰜉',
                            extra_information=ReadableInformation(
                                text="""Do you want to replace the current installed \
version of ubo-app with the selected version?""",
                            ),
                            color=INFO_COLOR,
                            dismiss_on_close=True,
                            display_type=NotificationDisplayType.STICKY,
                            show_dismiss_action=True,
                            actions=[
                                NotificationDispatchItem(
                                    icon='󰜉',
                                    store_action=UpdateManagerRequestUpdateAction(
                                        version=version,
                                    ),
                                ),
                            ],
                        ),
                    ),
                )
                for version in state.recent_versions
            ],
        ),
    )

    installed_versions_item = SubMenuItem(
        key='installed_versions',
        label='Installed versions',
        icon='󰯍',
        sub_menu=HeadlessMenu(
            title='󰯍Installed versions',
            items=[
                UboDispatchItem(
                    label=item.name,
                    **(
                        SELECTED_ITEM_PARAMETERS
                        if item.name == CURRENT_VERSION
                        else UNSELECTED_ITEM_PARAMETERS
                    ),
                    store_action=NotificationsAddAction(
                        notification=Notification(
                            id=f'update-manager-activate-{item.name}',
                            title=f'Activate {item.name} now?',
                            content=f'Press 󰯍 button to activate version "{item.name}"',
                            icon='󰯍',
                            extra_information=ReadableInformation(
                                text="""Do you want to activate the selected version of
    ubo-app?""",
                            ),
                            color=INFO_COLOR,
                            dismiss_on_close=True,
                            display_type=NotificationDisplayType.STICKY,
                            show_dismiss_action=True,
                            actions=[
                                NotificationActionItem(
                                    icon='󰯍',
                                    action=functools.partial(
                                        activate_version,
                                        version=item,
                                    ),
                                ),
                            ],
                        ),
                    ),
                )
                for item in get_installed_versions()
            ],
        ),
    )

    if state.update_status is UpdateStatus.CHECKING:
        items = [
            Item(
                label='Checking for updates...',
                icon='󰬬',
                background_color='#00000000',
            ),
        ]
    if state.update_status is UpdateStatus.FAILED_TO_CHECK:
        items = [
            UboDispatchItem(
                label='Failed to check for updates',
                store_action=UpdateManagerRequestCheckAction(),
                icon='󰜺',
                background_color=DANGER_COLOR,
            ),
        ]
    if state.update_status is UpdateStatus.UP_TO_DATE:
        items = [
            UboDispatchItem(
                label='Already up to date!',
                icon='󰄬',
                store_action=UpdateManagerRequestCheckAction(),
                background_color=SUCCESS_COLOR,
                color='#000000',
            ),
            recent_versions_item,
        ]
        if beta_versions:
            items.append(installed_versions_item)
    if state.update_status is UpdateStatus.OUTDATED:
        items = [
            *(
                []
                if state.latest_version is None
                else [
                    UboDispatchItem(
                        label=f'Update to v{state.latest_version}',
                        icon='󰬬',
                        store_action=NotificationsAddAction(
                            notification=Notification(
                                id=f'update-manager-latest-{state.latest_version}',
                                title=f'Install {state.latest_version} now?',
                                content='Press 󰬬 button to start installation of '
                                f'version "{state.latest_version}"',
                                icon='󰬬',
                                extra_information=ReadableInformation(
                                    text='Do you want to update to the latest version?',
                                ),
                                color=INFO_COLOR,
                                dismiss_on_close=True,
                                display_type=NotificationDisplayType.STICKY,
                                show_dismiss_action=True,
                                actions=[
                                    NotificationDispatchItem(
                                        icon='󰬬',
                                        store_action=UpdateManagerRequestUpdateAction(
                                            version=state.latest_version,
                                        ),
                                    ),
                                ],
                            ),
                        ),
                    ),
                ]
            ),
            recent_versions_item,
        ]
        if beta_versions:
            items.append(installed_versions_item)
    if state.update_status is UpdateStatus.UPDATING:
        items = [
            Item(
                label='Updating...',
                icon='󰇚',
                background_color='#00000000',
            ),
        ]

    return items



"""Implementation of a file system navigation application."""

from __future__ import annotations

import functools
import mimetypes
import stat
from dataclasses import replace
from pathlib import Path
from typing import TYPE_CHECKING

from kivy.utils import escape_markup
from redux import AutorunOptions
from ubo_gui.menu.types import ActionItem, HeadlessMenu, Item

from ubo_app.store.input.types import PathInputDescription
from ubo_app.store.main import store
from ubo_app.store.services.file_system import (
    FileSystemCopyAction,
    FileSystemEvent,
    FileSystemMoveAction,
    FileSystemRemoveAction,
    FileSystemReportSelectionAction,
    PathSelectorConfig,
)
from ubo_app.store.services.notifications import (
    Notification,
    NotificationActionItem,
    NotificationApplicationItem,
    NotificationDispatchItem,
    NotificationDisplayType,
    NotificationsAddAction,
)
from ubo_app.utils.async_ import create_task
from ubo_app.utils.error_handlers import report_service_error
from ubo_app.utils.file_system import human_readable_size
from ubo_app.utils.input import ubo_input

if TYPE_CHECKING:
    from collections.abc import Callable

FILE_VIEWER_SIZE_LIMIT = 2**11  # 2 KiB


def _file_info(path: Path) -> str:
    return f"""[b]Type:[/b] {
        'Directory'
        if path.is_dir()
        else 'Symlink'
        if path.is_symlink()
        else 'Block Device'
        if path.is_block_device()
        else 'Character Device'
        if path.is_char_device()
        else 'FIFO'
        if path.is_fifo()
        else 'Socket'
        if path.is_socket()
        else 'File'
    }
[b]Path:[/b] {escape_markup(path.as_posix())}
[b]Size:[/b] {'-' if path.is_dir() else human_readable_size(path.stat().st_size)}
[b]Owner:[/b] {path.owner()}
[b]Group:[/b] {path.group()}
[b]Permissions:[/b] {stat.filemode(path.stat().st_mode)}"""


def _get_file_content(path: Path) -> str:
    """Show the path in a notification."""
    try:
        content_bytes = path.read_bytes().replace(b'\0', b'\\x00')
        if len(content_bytes) > FILE_VIEWER_SIZE_LIMIT:
            content_bytes = (
                content_bytes[:FILE_VIEWER_SIZE_LIMIT]
                + (
                    f' [i][{len(content_bytes) - FILE_VIEWER_SIZE_LIMIT} more bytes]'
                    '[/i]'
                ).encode()
            )
    except Exception:  # noqa: BLE001
        report_service_error()
        return '[i][Error reading file content.][/i]'
    else:
        return (
            content_bytes.decode(errors='backslashreplace')
            .replace(
                ' ',
                '[color=#666]󱁐[/color]',
            )
            .replace(
                '\n',
                '[color=#666]󰌑[/color]\n',
            )
            .replace(
                '\t',
                '[color=#666][/color]',
            )
        )


def _copy(path: Path) -> None:
    """Copy the path to the clipboard."""

    async def act() -> None:
        destination, _ = await ubo_input(
            title='Copy Destination',
            prompt='Select the destination directory.]\n'
            f'[b]Source:[/b] {escape_markup(path.as_posix())}',
            descriptions=[
                PathInputDescription(
                    selector_config=PathSelectorConfig(
                        accepts_directories=True,
                        accepts_files=False,
                    ),
                ),
            ],
        )

        store.dispatch(
            FileSystemCopyAction(
                sources=[path],
                destination=Path(destination),
            ),
        )

    create_task(act())


def _move(path: Path) -> None:
    """Move the path to the clipboard."""

    async def act() -> None:
        destination, _ = await ubo_input(
            title='Move Destination',
            prompt='Select the destination directory.]\n'
            f'[b]Source:[/b] {escape_markup(path.as_posix())}',
            descriptions=[
                PathInputDescription(
                    selector_config=PathSelectorConfig(
                        accepts_directories=True,
                        accepts_files=False,
                    ),
                ),
            ],
        )

        store.dispatch(
            FileSystemMoveAction(
                sources=[path],
                destination=Path(destination),
            ),
        )

    create_task(act())


def _remove(path: Path) -> None:
    """Remove the path."""
    store.dispatch(
        NotificationsAddAction(
            notification=Notification(
                title='Confirm Removal',
                content='Are you sure you want to remove '
                f'"{escape_markup(path.as_posix())}"?',
                icon='󰆴',
                display_type=NotificationDisplayType.STICKY,
                dismiss_on_close=True,
                actions=[
                    NotificationDispatchItem(
                        key='confirm',
                        label='Remove',
                        icon='󰆴',
                        store_action=FileSystemRemoveAction(paths=[path]),
                        close_notification=True,
                    ),
                ],
            ),
        ),
    )


def _show_directory(path: Path) -> HeadlessMenu | None:
    """Show the path in a notification."""
    store.dispatch(
        NotificationsAddAction(
            notification=Notification(
                title=escape_markup(path.name),
                content=_file_info(path),
                icon='󰉋',
                display_type=NotificationDisplayType.STICKY,
                show_dismiss_action=False,
                actions=[
                    NotificationActionItem(
                        key='copy',
                        label='Copy Directory',
                        icon='󰆏',
                        action=functools.partial(_copy, path),
                        close_notification=False,
                    ),
                    NotificationActionItem(
                        key='move',
                        label='Move Directory',
                        icon='󰉒',
                        action=functools.partial(_move, path),
                        close_notification=False,
                    ),
                    NotificationActionItem(
                        key='remove',
                        label='Remove Directory',
                        icon='󰉘',
                        action=functools.partial(_remove, path),
                        close_notification=False,
                    ),
                ],
            ),
        ),
    )


def _show_file(path: Path) -> HeadlessMenu | None:
    """Show the path in a notification."""
    file_type, _ = mimetypes.guess_type(path)
    match file_type:
        case str(type_) if type_.startswith('image/'):
            from PIL import Image

            image = Image.open(path)
            width, height = image.size
            image_bytes = image.tobytes()
            view_action = NotificationApplicationItem(
                key='view',
                label='Open Image',
                icon='󰋩',
                application_id='ubo:raw-image-viewer',
                initialization_kwargs={
                    'image': image_bytes,
                    'width': width,
                    'height': height,
                },
                close_notification=False,
            )
        case _:
            view_action = NotificationApplicationItem(
                key='view',
                label='View File Content',
                icon='󰦪',
                application_id='ubo:raw-text-viewer',
                initialization_kwargs={
                    'text': _get_file_content(path),
                },
                close_notification=False,
            )

    store.dispatch(
        NotificationsAddAction(
            notification=Notification(
                title=escape_markup(path.name),
                content=_file_info(path),
                icon='󰈔',
                display_type=NotificationDisplayType.STICKY,
                show_dismiss_action=False,
                actions=[
                    view_action,
                    NotificationActionItem(
                        key='copy',
                        label='Copy File',
                        icon='󰆏',
                        action=functools.partial(_copy, path),
                        close_notification=False,
                    ),
                    NotificationActionItem(
                        key='move',
                        label='Move File',
                        icon='󰪹',
                        action=functools.partial(_move, path),
                        close_notification=False,
                    ),
                    NotificationActionItem(
                        key='remove',
                        label='Remove File',
                        icon='󰮘',
                        action=functools.partial(_remove, path),
                        close_notification=False,
                    ),
                ]
                if path.is_file()
                else [],
            ),
        ),
    )


def _select(path: Path) -> None:
    store.dispatch(FileSystemReportSelectionAction(path=path))


def _items_generator(config: PathSelectorConfig) -> Callable[[], list[ActionItem]]:
    path = config.initial_path or Path('/')
    if config.accepts_directories and config.accepts_files:
        select_directory = select_file = _select
    elif config.accepts_directories:
        select_directory = _select
        select_file = None
    elif config.accepts_files:
        select_directory = None
        select_file = _select
    else:
        select_directory = _show_directory
        select_file = _show_file

    @store.autorun(lambda _: None, options=AutorunOptions(memoization=False))
    def items(_: None) -> list[Item]:
        return (
            [
                ActionItem(
                    key='select',
                    label='[b]Select[/b]'
                    if config.accepts_directories
                    else '[b]Info[/b]',
                    icon='',
                    background_color='#2d5b86',
                    action=functools.partial(select_directory, path),
                ),
            ]
            if select_directory
            else []
        ) + [
            ActionItem(
                key=item.as_posix(),
                label=escape_markup(item.name),
                background_color='#303030' if select_directory is None else None,
                icon='󰉋',
                action=functools.partial(
                    open_path,
                    config=replace(config, initial_path=item),
                ),
            )
            if item.is_dir()
            else ActionItem(
                key=item.as_posix(),
                label=escape_markup(item.name),
                background_color='#303030'
                if item.is_file()
                and (
                    select_file is None
                    or (
                        config.acceptable_suffixes
                        and not any(
                            suffix in config.acceptable_suffixes
                            for suffix in item.suffixes
                        )
                    )
                )
                else None,
                icon='󰈔',
                action=functools.partial(select_file, item),
            )
            if select_file
            else Item(
                key=item.as_posix(),
                label=escape_markup(item.name),
                background_color='#303030',
                icon='󰈔',
            )
            for item in sorted(
                path.iterdir(),
                key=lambda x: x.name.lower(),
            )
            if config.show_hidden or not item.name.startswith('.')
        ]

    items()

    store.subscribe_event(FileSystemEvent, items)

    return items


def open_path(*, config: PathSelectorConfig | None = None) -> HeadlessMenu | None:
    """Open a directory and return a HeadlessMenu for its contents."""
    config = config or PathSelectorConfig()
    path = config.initial_path or Path('/')

    try:
        if path.is_dir():
            return HeadlessMenu(
                title=escape_markup(path.as_posix()),
                items=_items_generator(config),
            )
    except PermissionError:
        store.dispatch(
            NotificationsAddAction(
                notification=Notification(
                    title='Permission Denied',
                    content=f'Cannot access {escape_markup(path.as_posix())}.',
                    icon='󰍛',
                    display_type=NotificationDisplayType.FLASH,
                ),
            ),
        )
        return None
    except Exception:
        store.dispatch(
            NotificationsAddAction(
                notification=Notification(
                    title='Error',
                    content='An error occurred while accessing '
                    f'{escape_markup(path.as_posix())}',
                    icon='󰍛',
                    display_type=NotificationDisplayType.FLASH,
                ),
            ),
        )
        raise
    else:
        store.dispatch(
            NotificationsAddAction(
                notification=Notification(
                    title='Invalid Selection',
                    content=f'{escape_markup(path.as_posix())} is not a directory.',
                    icon='󰍛',
                    display_type=NotificationDisplayType.FLASH,
                ),
            ),
        )
        return None



"""Contains the serialization functions for the immutable objects."""

import base64
from typing import Any

import dill
from immutable import Immutable


def add_type_field(
    obj: Immutable,
    serialized: dict[str, Any],
) -> dict[str, Any]:
    """Add the type field to the serialized object."""
    return {
        **serialized,
        '_type': base64.b64encode(dill.dumps(obj.__class__)).decode('utf-8'),
    }



"""Reducer for infrared actions."""

from dataclasses import replace

from redux import (
    CompleteReducerResult,
    InitAction,
    InitializationActionError,
    ReducerResult,
)

from ubo_app.logger import logger
from ubo_app.store.services.assistant import (
    AssistantStartListeningAction,
    AssistantStopListeningAction,
    AssistantToggleListeningAction,
)
from ubo_app.store.services.infrared import (
    InfraredAction,
    InfraredHandleReceivedCodeAction,
    InfraredSendCodeAction,
    InfraredSendCodeEvent,
    InfraredSetShouldPropagateAction,
    InfraredSetShouldReceiveAction,
    InfraredState,
)
from ubo_app.store.services.keypad import (
    Key,
    KeypadAction,
    KeypadKeyPressAction,
    KeypadKeyReleaseAction,
)

KeyActionType = type[KeypadKeyPressAction] | type[KeypadKeyReleaseAction]

KEY_TO_INFRARED_CODES: dict[KeyActionType, dict[Key, tuple[str, str]]] = {
    KeypadKeyPressAction: {
        Key.L1: ('necx', '0xbf10'),
        Key.L2: ('necx', '0xbf11'),
        Key.L3: ('necx', '0xbf12'),
        Key.UP: ('nex', '0xbf05'),
        Key.DOWN: ('necx', '0xbf0d'),
        Key.BACK: ('necx', '0x1a'),
        Key.HOME: ('necx', '0x1c'),
    },
    KeypadKeyReleaseAction: {
        Key.L1: ('necx', '0x7076d'),
        Key.L2: ('necx', '0x70716'),
        Key.L3: ('necx', '0x70716'),
        Key.UP: ('necx', '0xbf06'),
        Key.DOWN: ('necx', '0x00feb05f'),
        Key.BACK: ('necx', '0xbf0e'),
        Key.HOME: ('necx', '0xbf09'),
    },
}
INFRARED_CODES_TO_KEY: dict[tuple[str, str], tuple[KeyActionType, Key]] = {
    (protocol, scancode): (action_type, key)
    for action_type in KEY_TO_INFRARED_CODES
    for key, (protocol, scancode) in KEY_TO_INFRARED_CODES[action_type].items()
}

# Define mappings for IR codes that should trigger Assistant actions
ASSISTANT_IR_CODES = {
    ('necx', '0xbf04'): AssistantStartListeningAction,
    ('necx', '0xbf06'): AssistantStopListeningAction,
    ('nec', '0xa01b'): AssistantToggleListeningAction,
}


def reducer(
    state: InfraredState | None,
    action: InfraredAction | KeypadAction,
) -> ReducerResult[
    InfraredState,
    InfraredAction | KeypadKeyPressAction | KeypadKeyReleaseAction,
    InfraredSendCodeEvent,
]:
    """Reducer for infrared actions."""
    if state is None:
        if isinstance(action, InitAction):
            return InfraredState()

        raise InitializationActionError(action)

    match action:
        case InfraredSendCodeAction():
            return CompleteReducerResult(
                state=state,
                events=[
                    InfraredSendCodeEvent(
                        protocol=action.protocol,
                        scancode=action.scancode,
                    ),
                ],
            )

        case InfraredSetShouldPropagateAction():
            return replace(
                state,
                should_propagate_keypad_actions=action.should_propagate,
            )

        case InfraredSetShouldReceiveAction():
            return replace(state, should_receive_keypad_actions=action.should_receive)

        case KeypadKeyPressAction() | KeypadKeyReleaseAction() if (
            state.should_propagate_keypad_actions
        ):
            return CompleteReducerResult(
                state=state,
                actions=[
                    InfraredSendCodeAction(
                        protocol=KEY_TO_INFRARED_CODES[type(action)][action.key][0],
                        scancode=KEY_TO_INFRARED_CODES[type(action)][action.key][1],
                    ),
                ],
            )

        case InfraredHandleReceivedCodeAction() if state.should_receive_keypad_actions:
            logger.info(
                'Received infrared code: %s, %s',
                action.protocol,
                action.scancode,
            )
            ir_code = (action.protocol, action.scancode)
            if ir_code in ASSISTANT_IR_CODES:
                action_class = ASSISTANT_IR_CODES[ir_code]
                logger.info(
                    'Triggered Assistant action: %s',
                    action_class,
                )
                return CompleteReducerResult(
                    state=state,
                    actions=[
                        action_class(),
                    ],
                )
            if ir_code not in INFRARED_CODES_TO_KEY:
                return state

            key_action_type, key = INFRARED_CODES_TO_KEY[
                (action.protocol, action.scancode)
            ]

            if key_action_type is KeypadKeyPressAction:
                return CompleteReducerResult(
                    state=state,
                    actions=[
                        KeypadKeyPressAction(key=key, pressed_keys={key}),
                    ],
                )
            if key_action_type is KeypadKeyReleaseAction:
                return CompleteReducerResult(
                    state=state,
                    actions=[
                        KeypadKeyReleaseAction(key=key, pressed_keys=set()),
                    ],
                )

            return state

        case _:
            return state



"""Definition of audio state, actions, and events."""

from __future__ import annotations

from dataclasses import field
from enum import StrEnum

from immutable import Immutable
from redux import BaseAction, BaseEvent

from ubo_app.utils.persistent_store import read_from_persistent_store


class AudioDevice(StrEnum):
    """Audio device enum."""

    INPUT = 'Input'
    OUTPUT = 'Output'


class AudioAction(BaseAction):
    """Audio action."""


class AudioInstallDriverAction(AudioAction):
    """Install audio driver action."""


class AudioSetVolumeAction(AudioAction):
    """Set volume action."""

    volume: float
    device: AudioDevice


class AudioChangeVolumeAction(AudioAction):
    """Change volume action."""

    amount: float
    device: AudioDevice


class AudioSetMuteStatusAction(AudioAction):
    """Set mute status action."""

    is_mute: bool
    device: AudioDevice


class AudioToggleMuteStatusAction(AudioAction):
    """Toggle mute status action."""

    device: AudioDevice


class AudioPlayChimeAction(AudioAction):
    """Play chime action."""

    name: str


class AudioPlayAudioSampleAction(AudioAction):
    """Play audio action."""

    sample: AudioSample


class AudioPlayAudioSequenceAction(AudioAction):
    """Play indexed audio action."""

    sample: AudioSample | None
    id: str
    index: int


class AudioSample(Immutable):
    """An audio sample."""

    data: bytes
    channels: int
    rate: int
    width: int


class AudioReportSampleAction(AudioAction):
    """Report audio sample action."""

    timestamp: float
    sample_speech_recognition: bytes
    sample: AudioSample


class AudioPlaybackDoneAction(AudioAction):
    """Playback done action."""

    id: str


class AudioStartRecordingAction(AudioAction):
    """Start recording action."""


class AudioStopRecordingAction(AudioAction):
    """Stop recording action."""


class AudioToggleRecordingAction(AudioAction):
    """toggle recording action."""


class AudioPlayRecordingAction(AudioAction):
    """Play recording action."""


class AudioEvent(BaseEvent):
    """Audio event."""


class AudioReportSampleEvent(AudioEvent):
    """Report audio sample event."""

    timestamp: float
    sample_speech_recognition: bytes
    sample: AudioSample


class AudioInstallDriverEvent(AudioEvent):
    """Install audio driver event."""


class AudioPlayChimeEvent(AudioEvent):
    """Play chime event."""

    name: str


class AudioPlayAudioSampleEvent(AudioEvent):
    """Play audio event."""

    volume: float
    sample: AudioSample


class AudioPlayAudioSequenceEvent(AudioEvent):
    """Play indexed audio event."""

    volume: float
    sample: AudioSample | None
    id: str
    index: int


class AudioPlaybackDoneEvent(AudioEvent):
    """Playback done event."""

    id: str


class AudioState(Immutable):
    """Audio state."""

    playback_volume: float = field(
        default=read_from_persistent_store(
            'audio_state:playback_volume',
            default=0.15,
        ),
    )
    is_playback_mute: bool = field(
        default=read_from_persistent_store(
            'audio_state:is_playback_mute',
            default=False,
        ),
    )
    capture_volume: float = field(
        default=read_from_persistent_store(
            'audio_state:capture_volume',
            default=0.5,
        ),
    )
    is_capture_mute: bool = field(
        default=read_from_persistent_store(
            'audio_state:is_capture_mute',
            default=False,
        ),
    )

    is_recording: bool = False
    recording: AudioSample | None = None



"""Provides the package handler for the system manager."""

from __future__ import annotations

import subprocess

from ubo_app.logger import get_logger
from ubo_app.utils.apt import install_package, uninstall_package

PACKAGE_WHITELIST = [
    'lightdm',
    'rpi-connect',
]

logger = get_logger('system-manager')


def package_handler(action: str, package: str) -> str:
    """Handle package actions."""
    if package not in PACKAGE_WHITELIST:
        return 'Package not in whitelist'

    try:
        if action == 'install':
            if package == 'lightdm':
                install_package('raspberrypi-ui-mods', force=True)
                subprocess.run(
                    [
                        '/usr/bin/env',
                        'sed',
                        '-i',
                        '/etc/lightdm/lightdm.conf',
                        '-e',
                        's|#\\?autologin-user=.*|autologin-user=ubo|',
                    ],
                    check=False,
                )
                subprocess.run(
                    [
                        '/usr/bin/env',
                        'raspi-config',
                        'nonint',
                        'do_wayland',
                        'W2',
                    ],
                    check=False,
                )
            else:
                install_package(package)
            return 'installed'
        if action == 'uninstall':
            uninstall_package(package)
            return 'uninstalled'
    except Exception:
        logger.exception('Failed to handle package action')
        return 'error'
    else:
        return 'Invalid package action'



"""provides a function to install and start Docker on the host machine."""

from __future__ import annotations

import subprocess
from pathlib import Path

from ubo_app.constants import USERNAME
from ubo_app.logger import get_logger

logger = get_logger('system-manager')


def _delete_composition(*args: str) -> str | None:
    """Delete a Docker composition by path."""
    logger.info(
        'Deleting composition:',
        extra={'composition_path': args[0] if args else 'unknown'},
    )
    # Delete a composition by full path
    if not args:
        logger.error('No composition path provided for deletion')
        return 'error: no composition path provided'

    composition_path = Path(args[0])
    # Security validation: ensure the path is within a docker_compositions directory
    if composition_path.parent.name != 'docker_compositions':
        logger.error(
            'Invalid composition path: not in docker_compositions directory',
            extra={
                'composition_path': str(composition_path),
                'parent': str(composition_path.parent),
            },
        )
        return 'error: path must be within docker_compositions directory'

    # Check if directory exists
    if not composition_path.exists():
        logger.warning(
            'Composition directory does not exist',
            extra={'composition_path': str(composition_path)},
        )
        return 'done'  # Already deleted, return success

    # Delete the directory (Docker containers create root-owned files)
    try:
        subprocess.run(  # noqa: S603
            [
                '/usr/bin/env',
                'rm',
                '-rf',
                str(composition_path),
            ],
            check=True,
        )
        logger.info(
            'Deleted composition',
            extra={'composition_path': str(composition_path)},
        )
    except subprocess.CalledProcessError:
        logger.exception(
            'Failed to delete composition',
            extra={'composition_path': str(composition_path)},
        )
        return 'error: failed to delete composition'

    return 'done'


def docker_handler(command: str, *args: str) -> str | None:
    """Install and start Docker on the host machine."""
    if command == 'install':
        try:
            process = subprocess.run(  # noqa: S603
                Path(__file__).parent.parent / 'scripts/install_docker.sh',
                env={'USERNAME': USERNAME},
                check=True,
            )
            process.check_returncode()
        except Exception:
            logger.exception('Error installing Docker')
            return 'error'
        else:
            return 'installed'

    if command == 'start':
        subprocess.run(
            [
                '/usr/bin/env',
                'systemctl',
                'start',
                'docker.socket',
            ],
            check=False,
        )
        subprocess.run(
            [
                '/usr/bin/env',
                'systemctl',
                'start',
                'docker.service',
            ],
            check=False,
        )
    elif command == 'stop':
        subprocess.run(
            [
                '/usr/bin/env',
                'systemctl',
                'stop',
                'docker.socket',
            ],
            check=False,
        )
        subprocess.run(
            [
                '/usr/bin/env',
                'systemctl',
                'stop',
                'docker.service',
            ],
            check=False,
        )
    elif command == 'composition_delete':
        return _delete_composition(*args)

    return 'done'



"""Background running mixin abstract base class."""

from __future__ import annotations

import abc
import threading
from typing import TYPE_CHECKING, final

from ubo_app.colors import DANGER_COLOR
from ubo_app.engines.abstraction.engine import EngineMixin
from ubo_app.logger import logger
from ubo_app.store.main import store
from ubo_app.store.services.notifications import Notification, NotificationsAddAction
from ubo_app.utils.async_ import create_task
from ubo_app.utils.error_handlers import report_service_error

if TYPE_CHECKING:
    import asyncio


class BackgroundRunningMixin(EngineMixin, abc.ABC):
    """Base class for third-party background running engines."""

    def __init__(self, *, label: str | None = None) -> None:
        """Initialize `BackgroundRunningMixin`."""
        self._task = None
        self._run_lock = threading.Lock()
        self._is_running: bool = False
        super().__init__(label=label)

    @final
    def _task_done_callback(self, task: asyncio.Task[None]) -> None:
        self._is_running = False
        self._task = None
        if not task.cancelled() and task.exception():
            logger.exception(
                'An error occurred while running the engine',
                extra={
                    'engine_name': self.name,
                },
                exc_info=task.exception(),
            )
            report_service_error(exception=task.exception())
            store.dispatch(
                NotificationsAddAction(
                    notification=Notification(
                        title=self.name,
                        content=(
                            f'An error occurred while running "{self.name}".'
                            'Please check the logs for more details.'
                        ),
                        color=DANGER_COLOR,
                    ),
                ),
            )

    @final
    def _set_task(self, task: asyncio.Task[None]) -> None:
        self._task = task
        self._task.add_done_callback(self._task_done_callback)

    @abc.abstractmethod
    async def _run(self) -> None:
        raise NotImplementedError

    def run(self) -> bool:
        """Run the engine if it is not already running."""
        with self._run_lock:
            if self._is_running:
                return True
            self._is_running = True
            create_task(
                self._run(),
                callback=self._set_task,
                name='VoskEngine.run',
            )
            return True

    @final
    def stop(self) -> None:
        """Stop the engine if it is running."""
        if not self._task:
            return
        self._task.cancel()

    def should_be_running(self) -> bool:
        """Check if the engine should be running."""
        return False

    @final
    def decide_running_state(self) -> None:
        """Decide whether the engine should be running based on the state."""
        if self.should_be_running():
            self.run()
        else:
            self.stop()



# pyright: reportMissingModuleSource=false
# ruff: noqa: D100, D103
from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, Any, TypeVar

from ubo_app.store.services.ethernet import NetState
from ubo_app.utils.bus_provider import get_system_bus

if TYPE_CHECKING:
    from asyncio.tasks import _FutureLike
    from collections.abc import Coroutine


T = TypeVar('T')


def wait_for(task: _FutureLike[T]) -> Coroutine[Any, Any, T]:
    return asyncio.wait_for(task, timeout=10.0)


from sdbus_async.networkmanager import (  # noqa: E402
    DeviceState,
    NetworkDeviceGeneric,
    NetworkManager,
)
from sdbus_async.networkmanager.enums import DeviceType  # noqa: E402

system_buses = {}


async def get_ethernet_device() -> NetworkDeviceGeneric | None:
    network_manager = NetworkManager(get_system_bus())
    devices_paths = await wait_for(network_manager.get_devices())
    for device_path in devices_paths:
        generic_device = NetworkDeviceGeneric(device_path, get_system_bus())
        if await wait_for(generic_device.device_type) == DeviceType.ETHERNET:
            return generic_device
    return None


async def get_ethernet_device_state() -> NetState:
    ethernet_device = await get_ethernet_device()
    if ethernet_device is None:
        return NetState.UNKNOWN

    state = await ethernet_device.state
    if state is DeviceState.UNKNOWN:
        return NetState.UNKNOWN
    if state in (
        DeviceState.DISCONNECTED,
        DeviceState.UNMANAGED,
        DeviceState.UNAVAILABLE,
        DeviceState.FAILED,
    ):
        return NetState.DISCONNECTED
    if state in (DeviceState.NEED_AUTH,):
        return NetState.NEEDS_ATTENTION
    if state in (
        DeviceState.DEACTIVATING,
        DeviceState.PREPARE,
        DeviceState.CONFIG,
        DeviceState.IP_CONFIG,
        DeviceState.IP_CHECK,
        DeviceState.SECONDARIES,
    ):
        return NetState.PENDING
    if state == DeviceState.ACTIVATED:
        return NetState.CONNECTED

    return NetState.UNKNOWN



"""Menus and actions for Docker images."""

from __future__ import annotations

from typing import TYPE_CHECKING

from docker_composition import check_composition
from docker_container import check_container
from docker_images import IMAGES
from docker_qrcode_page import DockerQRCodePage
from redux import AutorunOptions
from ubo_gui.menu.types import (
    ActionItem,
    HeadedMenu,
    HeadlessMenu,
    Item,
    SubMenuItem,
)

from ubo_app.colors import DANGER_COLOR
from ubo_app.store.main import store
from ubo_app.store.services.docker import (
    DockerImageFetchAction,
    DockerImageReleaseAction,
    DockerImageRemoveAction,
    DockerImageRemoveContainerAction,
    DockerImageRunAction,
    DockerImageStopAction,
    DockerItemStatus,
    ImageState,
)
from ubo_app.store.services.notifications import (
    Importance,
    Notification,
    NotificationActionItem,
    NotificationsAddAction,
)
from ubo_app.store.services.speech_synthesis import ReadableInformation
from ubo_app.store.ubo_actions import UboDispatchItem
from ubo_app.utils.async_ import create_task

if TYPE_CHECKING:
    from collections.abc import Callable, Sequence

    from ubo_gui.page import PageWidget

    from ubo_app.store.services.ip import IpNetworkInterface


def _show_delete_confirmation(image_id: str) -> None:
    """Show confirmation dialog before deleting a composition."""
    def _delete_action() -> None:
        """Execute the delete action."""
        store.dispatch(DockerImageRemoveAction(image=image_id))

    store.dispatch(
        NotificationsAddAction(
            notification=Notification(
                title='Delete Application?',
                content='All application data and files will be permanently deleted. '
                'This action cannot be undone.',
                importance=Importance.HIGH,
                icon='󰆴',
                actions=[
                    NotificationActionItem(
                        action=_delete_action,
                        icon='󰆴',
                        background_color=DANGER_COLOR,
                        label='Delete',
                    ),
                ],
            ),
        ),
    )


@store.with_state(lambda state: state.ip.interfaces if hasattr(state, 'ip') else None)
def image_menu(  # noqa: C901
    interfaces: Sequence[IpNetworkInterface] | None,
    image: ImageState,
) -> HeadedMenu:
    """Get the menu for the docker image."""
    ip_addresses = [
        ip for interface in interfaces or [] for ip in interface.ip_addresses
    ]
    items: list[Item] = []
    is_composition = image.id in IMAGES and IMAGES[image.id].is_composition

    def open_qrcode(port: str) -> Callable[[], PageWidget]:
        def action() -> PageWidget:
            return DockerQRCodePage(ips=ip_addresses, port=port)

        return action

    if image.status == DockerItemStatus.NOT_AVAILABLE:
        items.append(
            UboDispatchItem(
                label='Pull Images' if is_composition else 'Fetch',
                icon='󰇚',
                store_action=DockerImageFetchAction(image=image.id),
            ),
        )
    elif image.status == DockerItemStatus.FETCHING:
        pass
    elif image.status == DockerItemStatus.AVAILABLE:
        items.extend(
            [
                UboDispatchItem(
                    label='Start',
                    icon='󰐊',
                    store_action=DockerImageRunAction(image=image.id),
                ),
                ActionItem(
                    label='Delete Application' if is_composition else 'Remove Image',
                    icon='󰆴',
                    background_color=DANGER_COLOR,
                    action=lambda img_id=image.id: _show_delete_confirmation(img_id),
                ) if is_composition else UboDispatchItem(
                    label='Remove Image',
                    icon='󰆴',
                    store_action=DockerImageRemoveAction(image=image.id),
                    background_color=DANGER_COLOR,
                ),
            ],
        )
    elif image.status == DockerItemStatus.CREATED:
        items.extend(
            [
                UboDispatchItem(
                    label='Start',
                    icon='󰐊',
                    store_action=DockerImageRunAction(image=image.id),
                ),
                UboDispatchItem(
                    label='Release Resources' if is_composition else 'Remove Container',
                    icon='󰆴',
                    store_action=DockerImageReleaseAction(image=image.id)
                    if is_composition
                    else DockerImageRemoveContainerAction(image=image.id),
                    background_color=DANGER_COLOR if not is_composition else None,
                ),
            ],
        )
    elif image.status == DockerItemStatus.RUNNING:
        items.append(
            UboDispatchItem(
                label='Stop',
                key='stop',
                icon='󰓛',
                store_action=DockerImageStopAction(image=image.id),
            ),
        )
        if is_composition:
            items.append(
                UboDispatchItem(
                    label='Instructions',
                    key='instructions',
                    icon='󰋗',
                    store_action=NotificationsAddAction(
                        notification=Notification(
                            icon='󰋗',
                            title='Instructions',
                            content='',
                            extra_information=ReadableInformation(
                                text=image.instructions,
                            )
                            if image.instructions
                            else None,
                        ),
                    ),
                ),
            )
        else:
            items.append(
                SubMenuItem(
                    label='Ports',
                    key='ports',
                    icon='󰙜',
                    sub_menu=HeadlessMenu(
                        title='Ports',
                        items=[
                            ActionItem(
                                label=port,
                                key=port,
                                icon='󰙜',
                                action=open_qrcode(port.split(':')[-1]),
                            )
                            if port.startswith('0.0.0.0') and ip_addresses  # noqa: S104
                            else Item(label=port, icon='󰙜')
                            for port in image.ports
                        ],
                        placeholder='No ports',
                    ),
                ),
            )
    elif image.status == DockerItemStatus.PROCESSING:
        pass

    if is_composition:
        messages = {
            DockerItemStatus.NOT_AVAILABLE: 'Need to fetch images',
            DockerItemStatus.FETCHING: 'Images are being fetched',
            DockerItemStatus.AVAILABLE: 'Images are ready but composition is not '
            'running',
            DockerItemStatus.CREATED: 'Composition is created but not running',
            DockerItemStatus.RUNNING: 'Composition is running',
            DockerItemStatus.ERROR: 'We have an error, please check the logs',
            DockerItemStatus.PROCESSING: 'Waiting...',
        }
    else:
        # For containers, use note from IMAGES
        # For compositions, use generic message
        running_message = (
            IMAGES[image.id].note
            if image.id in IMAGES
            else 'Container is running'
        )
        messages = {
            DockerItemStatus.NOT_AVAILABLE: 'Need to fetch the image',
            DockerItemStatus.FETCHING: 'Image is being fetched',
            DockerItemStatus.AVAILABLE: 'Image is ready but container is not running',
            DockerItemStatus.CREATED: 'Container is created but not running',
            DockerItemStatus.RUNNING: running_message or 'Container is running',
            DockerItemStatus.ERROR: 'We have an error, please check the logs',
            DockerItemStatus.PROCESSING: 'Waiting...',
        }

    return HeadedMenu(
        title=f'Docker - {image.label}',
        heading=image.label,
        sub_heading=messages[image.status],
        items=items,
        placeholder='',
    )


def docker_item_menu(image_id: str) -> Callable[[], HeadedMenu]:
    """Get the menu items for the Docker service."""
    # Don't check status during ongoing operations (FETCHING, PROCESSING)
    # The operation itself manages the status
    def menu_with_check(image: ImageState) -> HeadedMenu:
        # Only check status if not in middle of an operation
        if image.status not in (DockerItemStatus.FETCHING, DockerItemStatus.PROCESSING):
            is_composition = image_id in IMAGES and IMAGES[image_id].is_composition
            if is_composition:
                create_task(check_composition(id=image_id))
            else:
                check_container(image_id=image_id)
        return image_menu(image)

    return store.autorun(
        lambda state: getattr(state.docker, image_id),
        lambda state: (
            getattr(state.docker, image_id),
            state.ip.interfaces if hasattr(state, 'ip') else None,
        ),
        options=AutorunOptions(default_value=None),
    )(menu_with_check)



# ruff: noqa: D100, D101, D102, D107
from __future__ import annotations

import asyncio
import pathlib
import re
import subprocess

from commands import install_service
from constants_ import CODE_BINARY_PATH
from kivy.clock import mainthread
from kivy.lang.builder import Builder
from kivy.properties import NumericProperty, StringProperty

from ubo_app.colors import DANGER_COLOR
from ubo_app.logger import logger
from ubo_app.store.core.types import CloseApplicationAction
from ubo_app.store.main import store
from ubo_app.store.services.notifications import (
    Chime,
    Notification,
    NotificationDisplayType,
    NotificationsAddAction,
)
from ubo_app.store.services.vscode import VSCodeLoginEvent
from ubo_app.utils.async_ import create_task
from ubo_app.utils.gui import UboPageWidget


class LoginPage(UboPageWidget):
    stage: int = NumericProperty(0)
    url: str | None = StringProperty()
    code: str | None = StringProperty()

    def __init__(
        self: LoginPage,
        *args: object,
        **kwargs: object,
    ) -> None:
        super().__init__(*args, **kwargs)
        store.subscribe_event(
            VSCodeLoginEvent,
            lambda: store.dispatch(
                CloseApplicationAction(application_instance_id=self.id),
            ),
        )
        create_task(self.login())

    async def login(self: LoginPage) -> None:
        try:
            self.process = await asyncio.create_subprocess_exec(
                CODE_BINARY_PATH.as_posix(),
                'tunnel',
                '--accept-server-license-terms',
                'user',
                'login',
                '--provider',
                'github',
                stdin=asyncio.subprocess.DEVNULL,
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.DEVNULL,
            )
            if self.process.stdout is None:
                return
            output = (await self.process.stdout.readline()).decode()
            regex = (
                r'To grant access to the server, please log into (?P<url>[^\s]*) and '
                r'use code (?P<code>[^\s]*)'
            )
            match = re.search(regex, output)
            if match:

                def set_properties() -> None:
                    self.url = match.group('url')
                    self.code = match.group('code')
                    self.stage = 1

                mainthread(set_properties)()
                await self.process.wait()
            else:
                logger.error(
                    'VSCode: Failed to login: invalid output',
                    extra={'output': output},
                )
                store.dispatch(
                    NotificationsAddAction(
                        notification=Notification(
                            id='vscode:login',
                            title='VSCode',
                            content='Failed to login: invalid output',
                            display_type=NotificationDisplayType.STICKY,
                            color=DANGER_COLOR,
                            icon='󰜺',
                            chime=Chime.FAILURE,
                        ),
                    ),
                )
        except subprocess.CalledProcessError:
            store.dispatch(
                NotificationsAddAction(
                    notification=Notification(
                        id='vscode:error:login',
                        title='VSCode',
                        content='Failed to login: process error',
                        display_type=NotificationDisplayType.STICKY,
                        color=DANGER_COLOR,
                        icon='󰜺',
                        chime=Chime.FAILURE,
                    ),
                ),
            )
            raise
        finally:
            await install_service()

    def on_close(self: LoginPage) -> None:
        self.process.kill() if self.process.returncode is None else None


Builder.load_file(
    pathlib.Path(__file__).parent.joinpath('login_page.kv').resolve().as_posix(),
)



"""Async remote store client for dispatching operations to a gRPC server."""

from __future__ import annotations

import asyncio
import logging
from typing import TYPE_CHECKING, cast, overload

from betterproto.lib.std.google import protobuf as betterproto_protobuf
from grpclib.client import Channel

import ubo_bindings.ubo.v1
from ubo_bindings.secrets.v1 import QuerySecretRequest, SecretsServiceStub
from ubo_bindings.store.v1 import (
    DispatchActionRequest,
    DispatchEventRequest,
    StoreServiceStub,
    SubscribeEventRequest,
    SubscribeStoreRequest,
)

if TYPE_CHECKING:
    from collections.abc import Callable

    from betterproto import Message

    from ubo_bindings.ubo.v1 import Action, Event


def _unpack_from_any(message: betterproto_protobuf.Any) -> Message:
    if not message.type_url.startswith('type.googleapis.com/'):
        msg = f'Unsupported type URL: {message.type_url}'
        raise ValueError(msg)

    type_name = message.type_url[len('type.googleapis.com/') :]

    if type_name.startswith('google.protobuf.'):
        type_name = type_name[len('google.protobuf.') :]
        cls = cast('type[Message]', getattr(betterproto_protobuf, type_name, None))
        if cls is not None:
            return cls.FromString(message.value)

    if type_name.startswith('ubo_bindings.ubo.v1.'):
        type_name = type_name[len('ubo_bindings.ubo.v1.') :]
        cls = cast('type[Message]', getattr(ubo_bindings.ubo.v1, type_name, None))
        if cls is not None:
            return cls.FromString(message.value)

    msg = f'Unknown type URL: {message.type_url}'
    raise ValueError(msg)


class UboRPCClient:
    """Async remote store for dispatching operations to a gRPC server."""

    def __init__(self, host: str, port: int) -> None:
        """Initialize the async remote store."""
        self.event_loop = asyncio.get_event_loop()
        self.channel = Channel(host=host, port=port, loop=self.event_loop)
        self.store_service = StoreServiceStub(self.channel)
        self.secrets_service = SecretsServiceStub(self.channel)

    def close(self) -> None:
        """Close the channel."""
        self.channel.close()

    @overload
    def dispatch(self, *, action: Action) -> None: ...
    @overload
    def dispatch(self, *, event: Event) -> None: ...
    def dispatch(
        self,
        *,
        action: Action | None = None,
        event: Event | None = None,
    ) -> None:
        """Dispatch an operation to the remote store."""
        if action is not None:
            self.event_loop.create_task(
                self.store_service.dispatch_action(
                    DispatchActionRequest(action=action),
                ),
            )
        if event is not None:
            self.event_loop.create_task(
                self.store_service.dispatch_event(DispatchEventRequest(event=event)),
            )

    def subscribe_event(
        self,
        event_type: Event,
        callback: Callable[[Event], None],
    ) -> Callable[[], None]:
        """Subscribe to the remote store."""

        async def iterator() -> None:
            async for response in self.store_service.subscribe_event(
                SubscribeEventRequest(event=event_type),
            ):
                callback(response.event)

        task = self.event_loop.create_task(iterator())

        def unsubscribe() -> None:
            task.cancel()

        return unsubscribe

    def autorun(
        self,
        selectors: list[str],
    ) -> Callable[[Callable[[list], None]], Callable[[], None]]:
        """Autorun a function based on store changes."""

        def wrapper(callback: Callable[[list], None]) -> Callable[[], None]:
            async def iterator() -> None:
                try:
                    async for response in self.store_service.subscribe_store(
                        SubscribeStoreRequest(selectors=selectors),
                    ):
                        try:
                            callback(
                                [_unpack_from_any(item) for item in response.results],
                            )
                        except Exception:
                            logging.getLogger().exception(
                                'Error in autorun callback',
                            )
                except Exception:
                    logging.getLogger().exception('Error in autorun subscription')

            task = self.event_loop.create_task(iterator())

            def unsubscribe() -> None:
                task.cancel()

            return unsubscribe

        return wrapper

    async def query_secret(
        self,
        key: str,
        *,
        covered: bool = False,
        default: str | None = None,
    ) -> str | None:
        """Query a secret from the secrets manager."""
        result = await self.secrets_service.query_secret(
            QuerySecretRequest(key=key, covered=covered),
        )

        match result.error:
            case 'Secret not found':
                return default
            case None:
                if result.value is None:
                    msg = 'Secret value is None and there is no error.'
                    raise RuntimeError(msg)
                return result.value
            case _:
                msg = f'Error querying secret: {result.error}'
                raise RuntimeError(msg)



"""handle audio commands."""

from __future__ import annotations

import subprocess
import time
from pathlib import Path

from ubo_app.logger import get_logger

DEVICE = '1-001a'
DRIVER_PATH = Path('/sys/bus/i2c/drivers/wm8960')


logger = get_logger('system-manager')


def audio_handler(command: str) -> str | None:
    """Handle audio commands."""
    if command == 'install':
        try:
            process = subprocess.run(  # noqa: S603
                Path(__file__).parent.parent / 'scripts/install_wm8960.sh',
                check=True,
            )
            process.check_returncode()
        except Exception:
            logger.exception('Error installing WM8960 driver')
            return 'error'
        else:
            return 'installed'

    if command == 'failure_report':
        logger.info('Audio failure report received, rebinding device...')
        try:
            (DRIVER_PATH / 'unbind').write_text(DEVICE)
            time.sleep(1)
            (DRIVER_PATH / 'bind').write_text(DEVICE)
        except Exception as e:
            logger.exception('Error rebinding device', exc_info=e)
            return 'error'
        else:
            logger.info('Device has been rebound.')
            return 'done'
    return None



"""Images to be used in Docker containers."""

from __future__ import annotations

import asyncio
import json
import os
import secrets as py_secrets
import string
from dataclasses import field
from pathlib import Path
from typing import TYPE_CHECKING, Any, cast

import aiohttp
from immutable import Immutable

from ubo_app.constants import (
    CONFIG_PATH,
    DEBUG_DOCKER,
    GRPC_ENVOY_LISTEN_PORT,
    GRPC_LISTEN_PORT,
)
from ubo_app.logger import logger
from ubo_app.store.input.types import QRCodeInputDescription, WebUIInputDescription
from ubo_app.store.services.speech_synthesis import ReadableInformation
from ubo_app.utils import IS_RPI, secrets
from ubo_app.utils.input import ubo_input

if TYPE_CHECKING:
    from collections.abc import Callable, Coroutine

ENVOY_TEMPLATE_PATH = Path(__file__).parent / 'assets' / 'envoy.yaml.tmpl'
ENVOY_CONFIG_PATH = Path(__file__).parent / 'assets' / 'envoy.yaml'
COMPOSITIONS_PATH = CONFIG_PATH / 'docker_compositions'


async def prepare_envoy() -> bool:
    """Prepare Envoy for gRPC."""
    process = await asyncio.create_subprocess_exec(
        '/usr/bin/env',
        'envsubst',
        env={
            'GRPC_ENVOY_LISTEN_PORT': f'{GRPC_ENVOY_LISTEN_PORT}',
            'GRPC_LISTEN_PORT': f'{GRPC_LISTEN_PORT}',
            'GRPC_SERVER_HOST': '127.0.0.1' if IS_RPI else 'host.docker.internal',
            **os.environ,
        },
        stdin=ENVOY_TEMPLATE_PATH.open(),
        stdout=ENVOY_CONFIG_PATH.open('w'),
    )
    await process.wait()
    return process.returncode == 0


async def prepare_immich() -> bool:
    """Prepare Immich for Docker Composition."""
    try:
        composition_id = 'immich'
        composition_path = COMPOSITIONS_PATH / composition_id
        logger.info(
            'Preparing Immich composition',
            extra={'composition_path': str(composition_path)},
        )
        composition_path.mkdir(exist_ok=True, parents=True)

        # Check if already set up
        if (composition_path / 'docker-compose.yml').exists() and (
            composition_path / '.env'
        ).exists():
            return True

        # Download docker-compose.yml
        async with aiohttp.ClientSession() as session:
            async with session.get(
                'https://github.com/immich-app/immich/releases/latest/download/docker-compose.yml',
            ) as response:
                response.raise_for_status()
                compose_content = await response.text()
                (composition_path / 'docker-compose.yml').write_text(compose_content)

            # Download .env template
            async with session.get(
                'https://github.com/immich-app/immich/releases/latest/download/example.env',
            ) as response:
                response.raise_for_status()
                env_content = await response.text()

        # Generate credentials
        creds = {
            'IMMICH_DB_PASSWORD': ''.join(
                py_secrets.choice(string.ascii_letters + string.digits)
                for _ in range(32)
            ),
            'IMMICH_DB_USERNAME': 'user_'
            + ''.join(
                py_secrets.choice(string.ascii_lowercase + string.digits)
                for _ in range(16)
            ),
        }

        # Save credentials
        for key, value in creds.items():
            secrets.write_secret(key=key, value=value)

        # Process .env content
        env_mappings = {
            'IMMICH_VERSION': 'release',
            'DB_DATABASE_NAME': 'immich',
            'TZ': os.environ.get('TZ', 'UTC'),
            'DB_PASSWORD': creds['IMMICH_DB_PASSWORD'],
            'DB_USERNAME': creds['IMMICH_DB_USERNAME'],
        }

        for key, value in env_mappings.items():
            env_content = env_content.replace(f'{key}=', f'{key}={value}\n# Original: ')
            env_content = env_content.replace(f'${{{key}}}', value)

        # Path replacements
        path_replacements = {
            './library': str(composition_path / 'library'),
            './postgres': str(composition_path / 'postgres'),
        }

        for rel, abs_path in path_replacements.items():
            env_content = env_content.replace(f'={rel}', f'={abs_path}')
            env_content = env_content.replace(rel, abs_path)

        (composition_path / '.env').write_text(env_content)

        # Create metadata.json
        metadata = {
            'label': 'Immich',
            'icon': '',
            'instructions': """Immich is installed and running!

Access the web interface at:
http://{{hostname}}:2283

On first visit, create an admin account to start uploading your photos and videos.""",
            'compose_id': 'immich',
        }
        (composition_path / 'metadata.json').write_text(json.dumps(metadata))

    except Exception:
        logger.exception('Failed to prepare Immich')
        return False
    else:
        return True


class ContainerEntry(Immutable):
    """Container entry."""

    id: str
    label: str
    icon: str
    path: str
    registry: str
    dependencies: list[str] | None = None
    ports: dict[str, int | list[int] | tuple[str, int] | None] = field(
        default_factory=dict,
    )
    hosts: dict[str, str] = field(default_factory=dict)
    hostname: str | None = None
    note: str | None = None
    environment_vairables: (
        dict[
            str,
            str
            | Coroutine[Any, Any, str]
            | Callable[[], str | Coroutine[Any, Any, str]],
        ]
        | None
    ) = None
    network_mode: str = 'bridge'
    dns: list[str] | None = None
    volumes: list[str] | None = None
    command: (
        str
        | list[str]
        | Callable[[], str | list[str] | Coroutine[Any, Any, str | list[str]]]
        | None
    ) = None
    prepare: Callable[[], Coroutine[Any, Any, bool] | bool] | None = None
    is_composition: bool = False

    @property
    def full_path(self) -> str:
        """Get full image path including registry if specified."""
        return f'{self.registry}/{self.path}'


IMAGES = {
    image.id: image
    for image in [
        ContainerEntry(
            id='home_assistant',
            label='Home Assistant',
            icon='󰟐',
            path='homeassistant/home-assistant:stable',
            registry='docker.io',
            ports={'8123/tcp': 8123},
        ),
        ContainerEntry(
            id='home_bridge',
            label='Home Bridge',
            icon='󰘘',
            path='homebridge/homebridge:latest',
            registry='docker.io',
        ),
        ContainerEntry(
            id='portainer',
            label='Portainer',
            icon='',
            path='portainer/portainer-ce:latest',
            registry='docker.io',
            volumes=['/var/run/docker.sock:/var/run/docker.sock'],
        ),
        ContainerEntry(
            id='pi_hole',
            label='Pi-hole',
            icon='󰇖',
            hostname='pi.hole',
            note='Password: admin',
            path='pihole/pihole:latest',
            ports={
                '53/tcp': 53,
                '53/udp': 53,
                '80/tcp': 80,
                '443/tcp': 443,
            },
            dns=['127.0.0.1', '1.1.1.1'],
            registry='docker.io',
        ),
        ContainerEntry(
            id='ollama',
            label='Ollama',
            icon='󰳆',
            path='ollama/ollama:latest',
            registry='docker.io',
            ports={'11434/tcp': 11434},
        ),
        ContainerEntry(
            id='open_webui',
            label='Open WebUI',
            icon='󰾔',
            path='open-webui/open-webui:main',
            registry='ghcr.io',
            dependencies=['ollama'],
            ports={'8080/tcp': 8080},
            hosts={'host.docker.internal': 'host-gateway'},
        ),
        ContainerEntry(
            id='ngrok',
            label='Ngrok',
            icon='󰛶',
            network_mode='host',
            path='ngrok/ngrok:latest',
            registry='docker.io',
            environment_vairables={
                'NGROK_AUTHTOKEN': lambda: ubo_input(
                    resolver=lambda code, _: code,
                    prompt='Enter the Ngrok Auth Token',
                    descriptions=[
                        QRCodeInputDescription(
                            instructions=ReadableInformation(
                                text="""\
Follow these steps:

1. Login to your ngrok account.
2. Get the authentication token from the dashboard.
3. Convert it to QR code.
4. Scan QR code to input the token.""",
                                picovoice_text="""\
Follow these steps:

1. Login to your {ngrok|EH N G EH R AA K} account
2. Get the authentication token from the dashboard
3. Convert it to {QR|K Y UW AA R} code
4. Scan QR code to input the token""",
                            ),
                            pattern=rf'^[a-zA-Z0-9]{20, 30}_[a-zA-Z0-9]{20, 30}$',
                        ),
                        WebUIInputDescription(),
                    ],
                ),
            },
            command=lambda: ubo_input(
                resolver=lambda code, _: code,
                prompt='Enter the command, for example: `http 80` or `tcp 22`',
                descriptions=[
                    QRCodeInputDescription(
                        instructions=ReadableInformation(
                            text='This is the command you would enter when running '
                            'ngrok. Refer to ngrok documentation for further '
                            'information',
                            picovoice_text="""\
This is the command you would enter when running {ngrok|EH N G EH R AA K}.
Refer to {ngrok|EH N G EH R AA K} documentation for further information""",
                        ),
                    ),
                    WebUIInputDescription(),
                ],
            ),
        ),
        ContainerEntry(
            id='immich',
            label='Immich',
            icon='',
            path='immich-app/immich-server:release',
            registry='ghcr.io',
            prepare=prepare_immich,
            is_composition=True,
            ports={'2283/tcp': 2283},
        ),
        *(
            [
                ContainerEntry(
                    id='alpine',
                    label='Alpine',
                    icon='',
                    path='alpine:latest',
                    registry='docker.io',
                ),
            ]
            if DEBUG_DOCKER
            else []
        ),
        ContainerEntry(
            id='envoy_grpc',
            label='Envoy for gRPC',
            icon='󱂇',
            path='thegrandpkizzle/envoy:1.26.1',
            prepare=prepare_envoy,
            command=['--config-path', 'envoy.yaml'],
            registry='docker.io',
            volumes=[
                f'{ENVOY_CONFIG_PATH}:/envoy.yaml',
            ],
            **cast(
                'Any',
                {'network_mode': 'host'}
                if IS_RPI
                else {
                    'ports': {
                        f'{GRPC_ENVOY_LISTEN_PORT}/tcp': GRPC_ENVOY_LISTEN_PORT,
                    },
                },
            ),
        ),
    ]
}



# ruff: noqa: D100, D103
from __future__ import annotations

from constants import ETHERNET_STATE_ICON_ID, ETHERNET_STATE_ICON_PRIORITY
from debouncer import DebounceOptions, debounce
from ethernet_manager import get_ethernet_device, get_ethernet_device_state

from ubo_app.store.main import store
from ubo_app.store.services.ethernet import NetState
from ubo_app.store.status_icons.types import StatusIconsRegisterAction
from ubo_app.utils.async_ import create_task


@debounce(
    wait=0.5,
    options=DebounceOptions(leading=True, trailing=True, time_window=0.6),
)
async def update_ethernet_icon() -> None:
    state = await get_ethernet_device_state()
    store.dispatch(
        StatusIconsRegisterAction(
            icon={
                NetState.CONNECTED: '󱊪',
                NetState.DISCONNECTED: '󰌙',
                NetState.PENDING: '󰌘',
                NetState.NEEDS_ATTENTION: '󰌚',
                NetState.UNKNOWN: '󰈅',
            }[state],
            priority=ETHERNET_STATE_ICON_PRIORITY,
            id=ETHERNET_STATE_ICON_ID,
        ),
    )


async def setup_listeners() -> None:
    ethernet_device = await get_ethernet_device()
    if not ethernet_device:
        return

    async for _ in ethernet_device.properties_changed:
        create_task(update_ethernet_icon())


def init_service() -> None:
    create_task(update_ethernet_icon())
    create_task(setup_listeners())
